from __future__ import print_function, division, absolute_import

import os
from shutil import copytree
import tempfile

from runner import main as runner
from differ import main as differ, SPLIT_MARKER
from test_utils import get_input_path, get_expected_path, generate_ttx_dump

TOOL = 'buildmasterotfs'
CMD = ['-t', TOOL]


# -----
# Tests
# -----

def test_cjk_var():
    """
    Builds all OTFs for the 'CJKVar' project and then diffs two of them.
    """
    input_dir = get_input_path('CJKVar')
    temp_dir = os.path.join(tempfile.mkdtemp(), 'CJKVar')
    copytree(input_dir, temp_dir)
    ds_path = os.path.join(temp_dir, 'CJKVar.designspace')
    runner(CMD + ['-o', '_{}'.format(ds_path)])

    otf1_path = os.path.join(
        temp_dir, 'Normal', 'Master_8', 'MasterSet_Kanji-w600.00.otf')
    otf2_path = os.path.join(
        temp_dir, 'Condensed', 'Master_8', 'MasterSet_Kanji_75-w600.00.otf')

    for otf_path in (otf1_path, otf2_path):
        actual_ttx = generate_ttx_dump(otf_path)
        expected_ttx = get_expected_path(
            os.path.basename(otf_path)[:-3] + 'ttx')
        assert differ([expected_ttx, actual_ttx,
                       '-s',
                       '<ttFont sfntVersion' + SPLIT_MARKER +
                       '    <checkSumAdjustment value=' + SPLIT_MARKER +
                       '    <created value=' + SPLIT_MARKER +
                       '    <modified value=',
                       '-r', r'^\s+Version.*;hotconv.*;makeotfexe'])



from __future__ import print_function

__copyright__ = """Copyright 2016 Adobe Systems Incorporated (http://www.adobe.com/). All Rights Reserved.
"""

__usage__ = """
autohint  AutoHinting program v1.6 Aug 28 2018
autohint -h
autohint -u
autohint -hfd
autohint -pfd
autohint [-g <glyph list>] [-gf <filename>] [-xg <glyph list>] [-xgf <filename>] [-cf path] [-a] [-logOnly] [-log <logFile path>] [-r] [-q] [-c] [-nf] [-ns] [-nb] [-wd] [-o <output font path>]  font-path

Auto-hinting program for PostScript and OpenType/CFF fonts.
"""

from afdko.ufotools import kProcessedGlyphsLayerName, kProcessedGlyphsLayer

__help__ = __usage__ + """

Takes a  list fonts, and an optional list of glyphs, and hints the
fonts. If the list of glyphs is supplied, the hinting is limited to the
specified glyphs.

Note that the hinting is better if the font's global alignment zones are
set usefully; at the very least, you should have entered values that
capture capital height, x height, ascender and descender heights, and
ascender and descender overshoot. The reports provided by the stemhist
tool are useful for choosing these.

By default, autothint will hint all glyphs in the font. Options allow you to
specify a sub-set of the glyphs for hinting.

Options:

-h	Print help

-u	Print usage

-hfd  Print a description of the format for defining a set of alternate
alignment zones in an "fontinfo" file.

-pfd  Print the default FDDict values for the source font: the
	alignment zone, stem width, and other global values. This is useful
	as a starting point for building FDDict defintions (see -hfd).

-pfdl Print the list of user-defined FDDict values, and which glyphs
	are included in each. This is useful for checking your FDDict
	definitions and glyph search terms. (see -hfd).

-g <glyphID1>,<glyphID2>,...,<glyphIDn>
	Hint only the specified list of glyphs. Note that all glyphs will be
	written to the output file. The list must be comma-delimited. The glyph
	references may be glyphID's, glyph names, or glyph CID's. If the latter,
	the CID value must be prefixed with the string "/". There must be no
	white-space in the glyph list.
	Examples:
		autohint -g A,B,C,68 myFont
		autohint -g /1030,/434,/1535,68 myCIDFont

	A range of glyphs may be specified by providing two names separated
	only by a hyphen:
		autohint -g zero-nine,onehalf myFont

	Note that the range will be resolved by expanding the glyph indices
	(GID)'s, not by alphabetic names.

-gf <file name>
	Hint only the list of glyphs contained in the specified file. The file
	must contain a comma-delimited list of glyph identifiers. Any number of
	space, tab, and new-line characters are permitted between glyph names
	and commas.

-xg, -xgf
	Same as -g and -gf, but will exclude the specified glyphs from hinting.

-cf <path>
	AC will try and add counter hints to only a short hard-coded list of
	glyphs:
		V counters:  "m", "M", "T", "ellipsis"
		H counters:  "element", "equivalence", "notelement", "divide".
	Counter hints help to keep the space between stems open and equal in size.

	To extend this list, use the option -cf followed by a path to a text file.
	The text file must contain one record per line. A  record references one
	glyph, and should consist of a single letter V or H, to indicate whether
	the counters should be vertical or horizontal, followd by a space or tab,
	followed by a glyph name. A maximum of 64 glyph names can be added to either
	the vertical or horizontal list. Example:
		V	ffi
		V	ffl
		V	f_f_j

	Alternatively, if there is a file named "fontinfo" in the same directory as
	the source font, this script will look in that file for lines with the format:
		VCounterChar ( <glyph name1>  <glyph name2> ... )
		HCounterChar ( <glyph name1>  <glyph name2> ... )
	and add the referenced glyphs to the counter hint lists.
	Example line:
		VCounterChar (ffi ffl f_f_j)

-logOnly
	Do not change any outlines, but report warnings for all selected
	glyphs, including those already hinted. The option -q is ignored.
	The option -a is implied.

-q	Quiet mode
	Will suppress comments from the auto-hinting library about recommended
	glyph outline changes.

-c	Permit changes to the glyph outline
	When this is turned on, the autohint program will fix a few issues:
	if there are many hint substitutions, it will try and shuffle the order
	of paths to reduce this, and it will flatten nearly straight curves.
	It no longer blunts sharp angles.That is better done with checkoutlines

-nf	Suppress generation of flex commands

-ns	Suppress hint substitution
	Do this only if you really want the smallest possible font file. This
	will use only one set of hints for the entire glyph.

-nb	Allow the font to have to no stem widths or blue values specified.
	Without this option, autohint will complain and quit.

-o <output font path>
	If not specified, autohint will write the hinted output to the original
	font path name.

-log	Write output to a log file

-all Hint all glyphs, even if the source glyphs is unchanged and the
	glyph has been hinted before. Applies only to UFO fonts.

-hf	Use history file.  Will create it if it does not already exist.
	Should not be used with UFO fonts, where another mechanism is used.

-a	Hint all glyphs that are in the history file, or are unhinted.
	Has effect only if the history file is being used.

-r	Re-hint glyphs. Glyphs not in the history file will be hinted even
	if they already have hints. However, glyphs will not be hinted if
	they both have not changed and are in the history file.

 -decimal  Use decimal coordinates, instead of rounding them to the
 nearest integer value.

-wd	write changed glyphs to default layer instead of '%s'

autohint can also apply different sets of alignment zones while hinting
a particular set of glyphs. This is useful for name-keyed fonts, which,
unlike CID fonts, only have one set of global alignment zones and stem
widths. By default, autohint uses the font's global alignment zones and
stem widths for each glyph. However, if there is a file named "fontinfo"
in the same directory as the input font file, autohint will check the
"fontinfo" file for definitions of alternate sets of alignment zones,
and the matching lists of glyphs to which they should be applied. To see
the format for these entries, use the option "-hfd". This allows one set
of glyphs to be hinted using a different set of zones and stem widths
than other glyphs. This isn't as useful as having real multiple hint
dictionaries in the font, as the final name-keyed font can only have one
set of alignment zones, but it does allow for improved hinting when
different sets of glyphs need different baselines.


autohint can maintain a history file, which allows you to avoid hinting glyphs
that have already been auto-hinted or manually hinted. When this is in use,
autohint will by default hint only  those glyphs that are not
already hinted, and also those glyphs which are hinted, but whose outlines have
changed since the last time autohint was run. autohint knows whether an outline
has changed by storing the outline in the history file whenever the glyph is
hinted, and then consulting the history file when it is asked to hint the glyph
again. By default, autohint does not maintain or use the history file, but this
can be turned on with an option.

When used, the history file is named "<PostScriptName>.plist", in the same
location as the parent font file. For each glyph, autohint stores a simplified
version of the outline coordinates. If this entry is missing for a glyph and the
glyph has hints, then autohint assumes it was manually hinted, and will by
default not hint it again. If the file is missing, autohint will assume that all
the glyphs were manually hinted, and you will have to use the option -a or -r to
hint any glyphs.
""" % (kProcessedGlyphsLayerName)


__FDDoc__ = """
By default, autohint uses the font's global alignment zones and stem
widths when hinting each glyph. However, if there is a file named
"fontinfo" in the same directory as the input font file, autohint will
check the "fontinfo" file for definitions of sets of alignment zones ( a
"FDDict"), and the matching lists of glyphs to which they should be
applied. This allows one set of glyphs to be hinted using a different
set of zones and stem widths than other glyphs. This isn't as useful as
having real multiple hint dictionaries in the font, as the final
name-keyed font can only have one set of alignment zones, but it does
allow for improved hinting when different sets of glyphs need different
alignment zones.

If FDDict definitions are used, then the global alignment zones and stem
widths in the source font will be ignored. For any glyphs not covered by
an explicit FDDict definition, autohint will synthesize an dummy FDDict,
where the zones are set outside of the the font's bounding box, so they
will not affect hinting. This is desirable for glyphs that have no
features that need to be aligned.

If autohint finds an FDDict named "FinalFont", then it will write that
set of values to the output font. Otherwise, it will merge all the
alignment zones and stem widths in the union of all the FDDict
definitions. If this merge fails because some of the alignment zones are
stem widths overlap, then you have to provide a "FinalFont" FDDict that
explicitly defines which stems and zones to use in the hinted output
font.

To use a dictionary of alignment zones and stem widths, you need to
define both the dictionary of alternate values, and the set of glyphs to
apply it to. The FDDict must be defined in the file before the set of
glyphs which belong to it. Both the FDDict and the glyph set define a
name; an FDDict is applied to the glyph set with the same name.


If you run autohint with the option "-pfd", it will print out the list
of FDDict values for the source font. You can use this text as a
starting point for your FDDict definitions.

You can also run autohint with the option "-pfdl". This will print the
user-defined FDDict defintions, and the list of glyphs associated with
each FDDict. Youc an use this to check your values, and to check which
glyphs are assigned to which FDDict. In particular, check the glyph list
for the first FDDict "No Alignment Zones": this list exists because
these glyphs did not match in the search terms for any user-defined
FDDict.

The definitions use the following syntax:

begin FDDict <name>
	<key-1> <value-1>
	<key-2> <value-2>
	...
	<key-n> <value-n>
end FDDict <name>

begin GlyphSet <name>
	<glyphname-1> <glyphname-2> ...
	<glyphname-n>
end GlyphSet <name>

The glyph names may be either a real glyph name, or a regular expression
designed to match several names.  A mini regex primer:
^      matches at the start of the glyph name
$      matches at the end
[aABb]  matches any one character in the set a, A, b, B.
[A-Z]  matches any one character in the set comprising the range from A-Z
[abA-Z]  matches any one character in the set comprising set set of a, b,
         and the characters in the range from A-Z
.      matches any single character
+      maches whatever preceded it one or more times
*      matches whatever preceded it none or more times.
\      An escape character that includes the following character without
       the second one being interpreted as a regex special character

Examples:
^[A-Z]$     Matches names with one character in the range from A-Z.
^[A-Z].+    Matches any name where the first character is in the range A-Z,
			and it is followed by one or more characters.
[A-Z].+    Matches any name with a character that is in the range A-Z and
           which is followed by one or more characters.
^[A-Z].*   Matches any name with one or more characters, and
            the first character is in the range A-Z
^.+\.smallcaps   Matches any name that contains ".smallcaps"
^.+\.smallcaps$   Matches any name that ends with ".smallcaps"
^.+\.s[0-24]0$     Matches any name that ends with ".s00",".s10",".s20", or ".s04"


Example FDDict and GlyphSet definitions.
********************************

begin FDDict ST_Smallcaps
	# I like to put the non hitn stuff first.
	OrigEmSqUnits 1000
	FontName AachenStd-Bold # This gets used as the hint dict name if the font is eventually built as a CID font.
	FlexOK true

	# Alignment zones. The first is a bottom zone, the rest ar top zones. See below.
	BaselineOvershoot -20
	BaselineYCoord 0
	CapHeight 900
	CapOvershoot 20
	LcHeight 700
	LcOvershoot 15

	# Stem widths.
	DominantV [236 267]
	DominantH [141 152]
end FDDict ST_Smallcaps


begin FDDict LM_Smallcaps
	OrigEmSqUnits 1000
	FontName AachenStd-Bold
	BaselineOvershoot -25
	BaselineYCoord 0
	CapHeight 950
	CapOvershoot 25
	LcHeight 750
	LcOvershoot 21
	DominantV [236 267]
	DominantH [141 152]
	FlexOK true
end FDDict LM_Smallcaps


begin GlyphSet LM_Smallcaps
	[Ll]\S+\.smallcap  [Mm]\S+\.smallcap
end GlyphSet LM_Smallcaps


begin GlyphSet ST_Smallcaps
	[Tt]\S+\.smallcap  [Ss]\S+\.smallcap
end GlyphSet ST_Smallcaps

********************************

Note that whitespace must exist between keywords and values, but is
otherwise ignored. "#" is a comment character: any occurrence of "#" and
all following text on a line is skipped. GlyphSet and FDDict definitions
may be intermixed, as long as any FDDict is defined before the GlyphSet
which refers to it.

You must provide at least two BlueValue pairs (the 'BaselineYCoord'
bottom zone and any top zone), and you must provide the DominantH and
DominantV keywords. All other keywords are optional.

The full set of recognized FDDict keywords are:

BlueValue pairs:
	# BaselineOvershoot is a bottom zone, the rest are top zones.
	BaselineYCoord
	BaselineOvershoot

	CapHeight
	CapOvershoot

	LcHeight
	LcOvershoot

	AscenderHeight
	AscenderOvershoot

	FigHeight
	FigOvershoot

	Height5
	Height5Overshoot

	Height6
	Height6Overshoot

OtherBlues pairs:
	# These
	Baseline5Overshoot
	Baseline5

	Baseline6Overshoot
	Baseline6

	SuperiorOvershoot
	SuperiorBaseline

	OrdinalOvershoot
	OrdinalBaseline

	DescenderOvershoot
	DescenderHeight

For zones which capture the bottom of a feature in the glyph, (BaselineYCoord and all the OtherBlues), the value
specifies the top of the zone, and the "Overshoot" is a negative value which specifes the offset to the bottom of the zone, e.g.
	BaselineYCoord 0
	BaselineOvershoot 12

For zones which capture the top of a feature in the glyph, (teh rest of the BlueValue zones), the value
specifies the bottom of the zone, and the "Overshoot" is a negapositivetive value which specifes the offset to the top of the zone, e.g.
	Height6 800
	Height6Overshoot 20


Note also that there is no implied sequential order of values. Height6 may have a value less than or equal to CapHeight.

The values for keywords in one FontDict definiton are completely independent of the values used in another FontDict. There is
no inheritance from one definition to the next.

All FontDicts must specify at least the BaselineYCoord and one top zone.

Miscellaneous values.
	FontName	PostScript font name. Only used by makeotf when building CID font.
	OrigEmSqUnits Single value: size of em-square. Only used by makeotf when building CID font.
	LanguageGroup 0 or 1. Specifies whether counter hints for ideographic glyphs should be applied.
	DominantV	List of dominant vertical stems, in the form  [ <stem-value-1> <stem-value-2> ...]
	DominantH	List of dominant horizontal  stems, in the form  [ <stem-value-1> <stem-value-2> ...]
	FlexOK		true or false.
	VCounterChars	List of characters to which counter hints may be applied, in the form [ <glyph-name-1> <glyph-name-2> ...]
	HCounterChars	List of characters to which counter hints may be applied, in the form [ <glyph-name-1> <glyph-name-2> ...]

Note for cognoscenti: the  autohint program code ignores StdHW and StdVW
entries if DominantV and DominantH entries are present, so I omit
writing the Std[HV]W keywords to fontinfo file. Also, autohint will add
any non-duplicate stem width values for StemSnap[HV] to the Dominant[HV]
stem width list, but the StemSnap[HV] entries are not necessary if the
full list of stem widths are supplied as values for the Dominant[HV]
keywords, hence I also  write the full stem list for the Dominant[HV]
keywords, and do not write the StemSnap[HV] keywords, to the fontinfo
file. Thsi is technically not right, as DominantHV array is supposed to
hold only two values, but the autohint program doesn't care, and I can
write fewer entries this way.
"""

# Methods:

# Parse args. If glyphlist is from file, read in entire file as single string,
# and remove all whitespace, then parse out glyph-names and GID's.

# For each font name:
#    use fontTools library to open font and extract CFF table.
#    If error, skip font and report error.
#    filter specified glyph list, if any, with list of glyphs in the font.
#    open font plist file, if any. If not, create empty font plist.
#    build alignment zone string
#    for identifier in glyph-list:
#        Get T2 charstring for glyph from parent font CFF table. If not present, report and skip.
#        get new alignment zone string if FD array index (which font dict is used) has changed.
#        Convert to bez time
#        Build autohint point list string, used to tell if glyph has been changed since the last time it was hinted.
#        If requested, check against plist dict, and skip if glyph is
#        already hinted or is manually hinted.
#        Call auto-hint library on bez string.
#        If change to the point list is permitted and happened, rebuild
#        autohint point list string.
#        Convert bez string to T2 charstring, and update parent font CFF.
#        add glyph hint entry to plist file
#    save font plist file.

import sys
import os
import re
import time
import tempfile
from fontTools.misc.py23 import open, tounicode
from fontTools.ttLib import TTFont, getTableModule
import plistlib
import warnings
from afdko import fdkutils, ufotools
from afdko.beztools import *
import traceback
import shutil

try:
    from psautohint import AUTOHINTEXE
except ImportError:
    AUTOHINTEXE = "autohintexe"

# warnings.simplefilter("ignore", RuntimeWarning) # supress waring about use of os.tempnam().

kACIDKey = "AutoHintKey"
NEWBEZ_SUFFIX = '.new'
gLogFile = None
kFontPlistSuffix  = ".plist"

class ACOptions:
	def __init__(self):
		self.inputPath = None
		self.outputPath = None
		self.glyphList = []
		self.excludeGlyphList = 0
		self.usePlistFile = 0
		self.hintAll = 0
		self.rehint = 0
		self.verbose = 1
		self.allowChanges = 0
		self.noFlex = 0
		self.noHintSub = 0
		self.allow_no_blues = 0
		self.hCounterGlyphs = []
		self.vCounterGlyphs = []
		self.counterHintFile = None
		self.logOnly = 0
		self.logFilePath = None
		self.printDefaultFDDict = 0
		self.printFDDictList = 0
		self.debug = 0
		self.allowDecimalCoords = 0
		self.writeToDefaultLayer = 0

class ACOptionParseError(Exception):
	pass

class ACFontInfoParseError(Exception):
	pass

class ACFontError(Exception):
	pass

class ACHintError(Exception):
	pass

class FDKEnvironmentError(Exception):
	pass

def logMsg(*args):
	for arg in args:
		msg = str(arg).strip()
		if not msg:
			print()
			sys.stdout.flush()
			if gLogFile:
				gLogFile.write(os.linesep)
				gLogFile.flush()
			return

		msg = re.sub(r"[\t]", "", msg)
		if msg[-1] == ",":
			msg = msg[:-1]
			print(msg, end='')
			sys.stdout.flush()
			if gLogFile:
				gLogFile.write(msg)
				gLogFile.flush()
		else:
			print(msg)
			sys.stdout.flush()
			if gLogFile:
				gLogFile.write(msg + os.linesep)
				gLogFile.flush()

def ACreport(*args):
	# long function used by the hinting library
	for arg in args:
		print(arg, end='')
	if arg[-1] != os.linesep:
		print()

def CheckEnvironment():
	txPath = 'tx'
	txError = 0
	command = "%s -u 2>&1" % (txPath)
	report = fdkutils.runShellCmd(command)
	if "options" not in report:
			txError = 1

	if  txError:
		logMsg("Please re-install the FDK. The path to the program 'tx' is not in the environment variable PATH.")
		raise FDKEnvironmentError

	command = '"%s" -u 2>&1' % AUTOHINTEXE
	report = fdkutils.runShellCmd(command)
	if "version" not in report:
		logMsg("Please re-install the FDK. The path to the program 'autohintexe' is not in the environment variable PATH.")
		raise FDKEnvironmentError

	return

global nameAliasDict
nameAliasDict = {}

def aliasName(glyphName):
	if nameAliasDict:
		alias = nameAliasDict.get(glyphName, glyphName)
		return alias
	return glyphName


def expandNames(glyphName):
	global nameAliasDict

	glyphRange = glyphName.split("-")
	if len(glyphRange) > 1:
		g1 = expandNames(glyphRange[0])
		g2 =  expandNames(glyphRange[1])
		glyphName =  "%s-%s" % (g1, g2)

	elif glyphName[0] == "/":
		glyphName = "cid" + glyphName[1:].zfill(5)
		if glyphName == "cid00000":
			glyphName = ".notdef"
			nameAliasDict[glyphName] = "cid00000"

	elif glyphName.startswith("cid") and (len(glyphName) < 8):
		glyphName =  "cid" + glyphName[3:].zfill(5)
		if glyphName == "cid00000":
			glyphName = ".notdef"
			nameAliasDict[glyphName] = "cid00000"

	return glyphName

def parseGlyphListArg(glyphString):
	glyphList = []
	for item in re.sub(r"[ \t\r\n,]+", ",", glyphString).split(","):
		name = expandNames(item)
		if name:
			glyphList.append(name)
	return glyphList


def parseCounterHintData(path):
	hCounterGlyphList = []
	vCounterGlyphList = []
	gf = open(path, "rt")
	data = gf.read()
	gf.close()
	lines = re.findall(r"([^\r\n]+)", data)
	# strip blank and comment lines
	lines = filter(lambda line: re.sub(r"#.+", "", line), lines)
	lines = filter(lambda line: line.strip(), lines)
	for line in lines:
		fields = line.split()
		if (len(fields) != 2) or (fields[0] not in ["V", "v", "H", "h"]) :
			print("\tError: could not process counter hint line '%s' in file %s. Doesn't look like V or H followed by a tab or space, and then a glyph name." % (line, path))
		elif  fields[0] in ["V", "v"]:
			vCounterGlyphList.append(fields[1])
		else:
			hCounterGlyphList.append(fields[1])
	return hCounterGlyphList, vCounterGlyphList



def checkFontinfoFile(options):
	# Check if there ia a makeotf fontinfo file in the input font directory. If so,
	# get any Vcounter or HCouunter glyphs from it.
	srcFontInfo = os.path.dirname(options.inputPath)
	srcFontInfo = os.path.join(srcFontInfo, "fontinfo")
	if os.path.exists(srcFontInfo):
		fi = open(srcFontInfo, "r")
		data = fi.read()
		fi.close()
		data = re.sub(r"#[^\r\n]+", "", data)
		counterGlyphLists = re.findall(r"([VH])CounterChars\s+\(\s*([^\)\r\n]+)\)", data)
		for entry in counterGlyphLists:
			glyphList = entry[1].split()
			if glyphList:
				if entry[0] == "V":
					options.vCounterGlyphs.extend(glyphList)
				else:
					options.hCounterGlyphs.extend(glyphList)

		if 	options.vCounterGlyphs or options.hCounterGlyphs:
			options.counterHintFile = srcFontInfo

def getOptions():
	global gLogFile
	options = ACOptions()
	i = 1
	numOptions = len(sys.argv)
	while i < numOptions:
		arg = sys.argv[i]
		if options.inputPath:
			raise ACOptionParseError("Option Error: All options must preceed the  input font path <%s>." % arg)

		if arg == "-h":
			print(__help__)
			command = '"%s" -v' % AUTOHINTEXE
			report = fdkutils.runShellCmd(command)
			logMsg( report)
			raise ACOptionParseError
		elif arg == "-u":
			print(__usage__)
			command = '"%s" -v' % AUTOHINTEXE
			report = fdkutils.runShellCmd(command)
			logMsg( report)
			raise ACOptionParseError
		elif arg == "-hfd":
			print(__FDDoc__)
			raise ACOptionParseError
		elif arg == "-pfd":
			options.printDefaultFDDict = 1
		elif arg == "-pfdl":
			options.printFDDictList = 1
		elif arg == "-hf":
			options.usePlistFile = 1
		elif arg == "-a":
			options.hintAll = 1
		elif arg == "-all":
			options.hintAll = 1
		elif arg == "-r":
			options.rehint = 1
		elif arg == "-q":
			options.verbose = 0
		elif arg == "-c":
			options.allowChanges = 1
		elif arg == "-nf":
			options.noFlex = 1
		elif arg == "-ns":
			options.noHintSub = 1
		elif arg == "-nb":
			options.allow_no_blues = 1
		elif arg in ["-xg", "-g"]:
			if arg == "-xg":
				options.excludeGlyphList = 1
			i = i +1
			glyphString = sys.argv[i]
			if glyphString[0] == "-":
				raise ACOptionParseError("Option Error: it looks like the first item in the glyph list following '-g' is another option.")
			options.glyphList += parseGlyphListArg(glyphString)
		elif arg in ["-xgf", "-gf"]:
			if arg == "-xgf":
				options.excludeGlyphList = 1
			i = i +1
			filePath = sys.argv[i]
			if filePath[0] == "-":
				raise ACOptionParseError("Option Error: it looks like the the glyph list file following '-gf' is another option.")
			try:
				gf = open(filePath, "rt")
				glyphString = gf.read()
				gf.close()
			except (IOError,OSError):
				raise ACOptionParseError("Option Error: could not open glyph list file <%s>." %  filePath)
			options.glyphList += parseGlyphListArg(glyphString)
		elif arg == "-cf":
			i = i +1
			filePath = sys.argv[i]
			if filePath[0] == "-":
				raise ACOptionParseError("Option Error: it looks like the the counter hint glyph list file following '-cf' is another option.")
			try:
				options.counterHintFile = filePath
				options.hCounterGlyphs, options.vCounterGlyphs = parseCounterHintData(filePath)
			except (IOError,OSError):
				raise ACOptionParseError("Option Error: could not open counter hint glyph list  file <%s>." %  filePath)
		elif arg == "-logOnly":
			options.logOnly = 1
		elif arg == "-log":
			i = i +1
			options.logFilePath = sys.argv[i]
			gLogFile = open(options.logFilePath, "wt")
		elif arg == "-o":
			i = i +1
			options.outputPath = sys.argv[i]
		elif arg == "-d":
			options.debug = 1
		elif arg in ["-decimal", "-dec"]:
			options.allowDecimalCoords = True
		elif arg =="-wd":
			options.writeToDefaultLayer = 1
		elif arg[0] == "-":
			raise ACOptionParseError("Option Error: Unknown option <%s>." %  arg)
		else:
			options.inputPath = arg
		i  += 1
	if not options.inputPath:
		raise ACOptionParseError("Option Error: You must provide a font file path.")

	if not os.path.exists(options.inputPath):
		raise ACOptionParseError("Option Error: The input font file path %s' does not exist." % (options.inputPath))
	else:
		options.inputPath = options.inputPath.rstrip(os.sep) # might be a UFO font. auto completion in some shells adds a dir separator, which then causes problems with os.path.dirname().

	checkFontinfoFile(options)

	if options.logOnly:
		options.verbose = 1
		options.hintAll = 1

	return options


def getGlyphID(glyphTag, fontGlyphList):
	glyphID = None
	try:
		glyphID = int(glyphTag)
		glyphName = fontGlyphList[glyphID]
	except IndexError:
		pass
	except ValueError:
		try:
			glyphID = fontGlyphList.index(glyphTag)
		except IndexError:
			pass
		except ValueError:
			pass
	return glyphID


def getGlyphNames(glyphTag, fontGlyphList, fontFileName):
	glyphNameList = []
	rangeList = glyphTag.split("-")
	prevGID = getGlyphID(rangeList[0], fontGlyphList)
	if prevGID == None:
		if len(rangeList) > 1:
			logMsg( "\tWarning: glyph ID <%s> in range %s from glyph selection list option is not in font. <%s>." % (rangeList[0], glyphTag, fontFileName))
		else:
			logMsg( "\tWarning: glyph ID <%s> from glyph selection list option is not in font. <%s>." % (rangeList[0], fontFileName))
		return None
	glyphNameList.append(fontGlyphList[prevGID])

	for glyphTag2 in rangeList[1:]:
		gid = getGlyphID(glyphTag2, fontGlyphList)
		if gid == None:
			logMsg( "\tWarning: glyph ID <%s> in range %s from glyph selection list option is not in font. <%s>." % (glyphTag2, glyphTag, fontFileName))
			return None
		for i in range(prevGID+1, gid+1):
			glyphNameList.append(fontGlyphList[i])
		prevGID = gid

	return glyphNameList


def filterGlyphList(options, fontGlyphList, fontFileName):
	# Return the list of glyphs which are in the intersection of the argument list and the glyphs in the font
	# Complain about glyphs in the argument list which are not in the font.
	if not options.glyphList:
		glyphList = fontGlyphList
	else:
		# expand ranges:
		glyphList = []
		for glyphTag in options.glyphList:
			glyphNames = getGlyphNames(glyphTag, fontGlyphList, fontFileName)
			if glyphNames != None:
				glyphList.extend(glyphNames)
		if options.excludeGlyphList:
			glyphList = [name for name in fontGlyphList if name not in glyphList[:]]
	return glyphList


def  openFontPlistFile(psName, dirPath):
	# Find or create the plist file. This hold a Python dictionary in repr() form,
	# key: glyph name, value: outline point list
	# This is used to determine which glyphs are manually hinted, and which have changed since the last
	# hint pass.
	fontPlist = None
	filePath = None
	isNewPlistFile = 1
	pPath1 = os.path.join(dirPath, psName + kFontPlistSuffix)
	if  os.path.exists(pPath1):
		filePath = pPath1
	else: # Crude approach to file length limitations. Since Adobe keeps face info in separate directories, I don't worry about name collisions.
		pPath2 = os.path.join(dirPath, psName[:-len(kFontPlistSuffix)] + kFontPlistSuffix)
		if os.path.exists(pPath2):
			filePath = pPath2
	if not filePath:
		filePath = pPath1
	else:
		try:
			fontPlist = plistlib.Plist.fromFile(filePath)
			isNewPlistFile = 0
		except (IOError, OSError):
			raise ACFontError("\tError: font plist file exists, but coud not be read <%s>." % filePath)
		except:
			raise ACFontError("\tError: font plist file exists, but coud not be parsed <%s>." % filePath)

	if  fontPlist == None:
		fontPlist =  plistlib.Plist()
	if not fontPlist.has_key(kACIDKey):
		fontPlist[kACIDKey] = {}
	return fontPlist, filePath, isNewPlistFile


fontInfoKeywordList = [
'FontName', #string
'OrigEmSqUnits',
'LanguageGroup',
'DominantV', #array
'DominantH', #array
'FlexOK', #string
'BlueFuzz',
'VCounterChars', #counter
'HCounterChars', #counter
'BaselineYCoord',
'BaselineOvershoot',
'CapHeight',
'CapOvershoot',
'LcHeight',
'LcOvershoot',
'AscenderHeight',
'AscenderOvershoot',
'FigHeight',
'FigOvershoot',
'Height5',
'Height5Overshoot',
'Height6',
'Height6Overshoot',
'DescenderOvershoot',
'DescenderHeight',
'SuperiorOvershoot',
'SuperiorBaseline',
'OrdinalOvershoot',
'OrdinalBaseline',
'Baseline5Overshoot',
'Baseline5',
'Baseline6Overshoot',
'Baseline6',
]

integerPattern = """ -?\d+"""
arrayPattern = """ \[[ ,0-9]+\]"""
stringPattern = """ \S+"""
counterPattern = """ \([\S ]+\)"""

def parseFontInfoString(fontInfoString):
	for item in fontInfoKeywordList:
		if item in ['FontName', 'FlexOK']:
			matchingExp = item + stringPattern
		elif item in ['VCounterChars', 'HCounterChars']:
			matchingExp = item + counterPattern
		elif item in ['DominantV', 'DominantH']:
			matchingExp = item + arrayPattern
		else:
			matchingExp = item + integerPattern

		try:
			print('\t%s' % re.search(matchingExp, fontInfoString).group())
		except:
			pass

flexPatthern = re.compile(r"preflx1[^f]+preflx2[\r\n](-*\d+\s+-*\d+\s+-*\d+\s+-*\d+\s+-*\d+\s+-*\d+\s+)(-*\d+\s+-*\d+\s+-*\d+\s+-*\d+\s+-*\d+\s+-*\d+\s+).+?flx([\r\n])",  re.DOTALL)
commentPattern = re.compile(r"[^\r\n]*%[^\r\n]*[\r\n]")
hintGroupPattern = re.compile(r"beginsubr.+?newcolors[\r\n]", re.DOTALL)
whiteSpacePattern = re.compile(r"\s+", re.DOTALL)



def makeACIdentifier(bezText):
	# Get rid of all the hint operators and their args
	# collapse flex to just the two rct's
	bezText = commentPattern.sub("", bezText)
	bezText = hintGroupPattern.sub("", bezText)
	bezText = flexPatthern.sub( "\1 rct\3\2 rct\3", bezText)
	bezText = whiteSpacePattern.sub("", bezText)
	return bezText


def openFile(path, outFilePath=None, useHashMap=False):
	if os.path.isfile(path):
		font =  openOpenTypeFile(path, outFilePath)
	else:
		# maybe it is a a UFO font.
		# We always use the hash map to skip glyphs that have been previously processed, unless the user has report only, not make changes.
		font =  openUFOFile(path, outFilePath, useHashMap)
	return font


def openUFOFile(path, outFilePath, useHashMap):
	# Check if has glyphs/contents.plist
	contentsPath = os.path.join(path, "glyphs", "contents.plist")
	if not os.path.exists(contentsPath):
		msg = "Font file must be a PS, CFF, OTF, or ufo font file: %s." % (path)
		logMsg(msg)
		raise ACFontError(msg)

	# If user has specified a path other than the source font path, then copy the entire
	# ufo font, and operate on the copy.
	if (outFilePath != None) and (os.path.abspath(path) != os.path.abspath(outFilePath)):
		msg = "Copying from source UFO font to output UFO font before processing..."
		logMsg(msg)
		if os.path.exists(outFilePath):
			shutil.rmtree(outFilePath)
		shutil.copytree(path , outFilePath)
		path = outFilePath
	font = ufotools.UFOFontData(path, useHashMap, ufotools.kAutohintName)
	font.useProcessedLayer = True
	font.requiredHistory.append(ufotools.kCheckOutlineName) # Programs in this list must be run before autohint, if the outlines have been edited.
	return font


def openOpenTypeFile(path, outFilePath):
	# If input font is  CFF or PS, build a dummy ttFont in memory..
	# return ttFont, and flag if is a real OTF font Return flag is 0 if OTF, 1 if CFF, and 2 if PS/
	fontType  = 0 # OTF
	tempPathCFF = fdkutils.get_temp_file_path()
	try:
		with open(path, "rb") as ff:
			head = ff.read(4)
	except (IOError, OSError):
		logMsg("Failed to open and read font file %s." % path)

	if head == b"OTTO": # it is an OTF font, can process file directly
		try:
			ttFont = TTFont(path)
		except (IOError, OSError):
			raise ACFontError("Error opening or reading from font file <%s>." % path)
		except TTLibError:
			raise ACFontError("Error parsing font file <%s>." % path)

		try:
			cffTable = ttFont["CFF "]
		except KeyError:
			raise ACFontError("Error: font is not a CFF font <%s>." % fontFileName)

	else:
		# It is not an OTF file.
		if head[0:2] == b'\x01\x00': # CFF file
			fontType = 1
			tempPathCFF = path
		else:  # It is a PS file. Convert to CFF.
			fontType =  2
			print("Converting Type1 font to temp CFF font file...")
			command="tx  -cff +b -std \"%s\" \"%s\" 2>&1" % (path, tempPathCFF)
			report = fdkutils.runShellCmd(command)
			if "fatal" in report:
				logMsg("Attempted to convert font %s  from PS to a temporary CFF data file." % path)
				logMsg(report)
				raise ACFontError("Failed to convert PS font %s to a temp CFF font." % path)

		# now package the CFF font as an OTF font.
		with open(tempPathCFF, "rb") as ff:
			data = ff.read()
		try:
			ttFont = TTFont()
			cffModule = getTableModule('CFF ')
			cffTable = cffModule.table_C_F_F_('CFF ')
			ttFont['CFF '] = cffTable
			cffTable.decompile(data, ttFont)
		except:
			logMsg( "\t%s" %(traceback.format_exception_only(sys.exc_info()[0], sys.exc_info()[1])[-1]))
			logMsg("Attempted to read font %s  as CFF." % path)
			raise ACFontError("Error parsing font file <%s>." % path)

	fontData = CFFFontData(ttFont, path, outFilePath, fontType, logMsg)
	return fontData


def cmpFDDictEntries(entry1, entry2):
	# entry = [glyphName, [fdIndex, glyphListIndex] ]
	if entry1[1][1] > entry2[1][1]:
		return 1
	elif entry1[1][1] < entry2[1][1]:
		return -1
	else:
		return 0

def hintFile(options):

	path = options.inputPath
	fontFileName = os.path.basename(path)
	logMsg("Hinting font %s. Start time: %s." % (path, time.asctime()))

	try:
		useHashMap = not options.logOnly # for UFO fonts only. We always use the hash map, unless the user has said to only report issues.
		fontData = openFile(path, options.outputPath, useHashMap)
		fontData.allowDecimalCoords = options.allowDecimalCoords
		if options.writeToDefaultLayer and hasattr(fontData, "setWriteToDefault"): # UFO fonts only
			fontData.setWriteToDefault()
	except (IOError, OSError):
		logMsg( traceback.format_exception_only(sys.exc_info()[0], sys.exc_info()[1])[-1])
		raise ACFontError("Error opening or reading from font file <%s>." % fontFileName)
	except:
		logMsg( traceback.format_exception_only(sys.exc_info()[0], sys.exc_info()[1])[-1])
		raise ACFontError("Error parsing font file <%s>." % fontFileName)

	#   filter specified list, if any, with font list.
	fontGlyphList = fontData.getGlyphList()
	glyphList = filterGlyphList(options, fontGlyphList, fontFileName)
	if not glyphList:
		raise ACFontError("Error: selected glyph list is empty for font <%s>." % fontFileName)

	# temp file names for input and output bez files, and for the fontinfo file.
	tempBez = fdkutils.get_temp_file_path()
	tempBezNew = '{}{}'.format(tempBez, NEWBEZ_SUFFIX)
	tempFI = fdkutils.get_temp_file_path()

	psName = fontData.getPSName()

	if (not options.logOnly) and options.usePlistFile:
		fontPlist, fontPlistFilePath, isNewPlistFile = openFontPlistFile(psName, os.path.dirname(path))
		if isNewPlistFile and  not (options.hintAll or options.rehint):
			logMsg("No hint info plist file was found, so all glyphs are unknown to autohint. To hint all glyphs, run autohint again with option -a to hint all glyphs unconditionally.")
			logMsg("Done with font %s. End time: %s." % (path, time.asctime()))
			fontData.close()
			return

	# Check counter glyphs, if any.
	if options.hCounterGlyphs or options.vCounterGlyphs:
		missingList = filter(lambda name: name not in fontGlyphList, options.hCounterGlyphs + options.vCounterGlyphs)
		if missingList:
			logMsg( "\tError: glyph named in counter hint list file '%s' are not in font: %s" % (options.counterHintFile, missingList) )

	#    build alignment zone string
	if (options.printDefaultFDDict):
		logMsg("Showing default FDDict Values:")
		fdDict = fontData.getFontInfo(psName, path, options.allow_no_blues, options.noFlex, options.vCounterGlyphs, options.hCounterGlyphs)
		parseFontInfoString(str(fdDict))
		fontData.close()
		return

	fdGlyphDict, fontDictList = fontData.getfdInfo(psName, path, options.allow_no_blues, options.noFlex, options.vCounterGlyphs, options.hCounterGlyphs, glyphList)

	if options.printFDDictList:
		# Print the user defined FontDicts, and exit.
		if fdGlyphDict:
			logMsg("Showing user-defined FontDict Values:")
			for fi in range(len(fontDictList)):
				fontDict = fontDictList[fi]
				logMsg("")
				logMsg(fontDict.DictName)
				parseFontInfoString(str(fontDict))
				gnameList = []
				itemList = fdGlyphDict.items()
				itemList.sort(cmpFDDictEntries)
				for gName, entry in itemList:
					if entry[0] == fi:
						gnameList.append(gName)
				logMsg("%d glyphs:" % len(gnameList))
				if len(gnameList) > 0:
					gTxt = " ".join(gnameList)
				else:
					gTxt = "None"
				logMsg(gTxt)
		else:
			logMsg("There are no user-defined FontDict Values.")
		fontData.close()
		return

	if fdGlyphDict == None:
		fdDict = fontDictList[0]
		with open(tempFI, "w") as fp:
			fp.write(tounicode(fdDict.getFontInfo()))
	else:
		if not options.verbose:
			logMsg("Note: Using alternate FDDict global values from fontinfo file for some glyphs. Remove option '-q' to see which dict is used for which glyphs.")


	#    for identifier in glyph-list:
	# 	Get charstring.
	removeHints = 1
	isCID = fontData.isCID()
	lastFDIndex = None
	reportCB = ACreport
	anyGlyphChanged = 0
	pListChanged = 0
	if isCID:
		options.noFlex = 1

	if options.verbose:
		verboseArg = ""
	else:
		verboseArg = " -q"
		dotCount = 0
		curTime = time.time()

	if options.allowChanges:
		suppressEditArg = ""
	else:
		suppressEditArg = " -e"

	if options.noHintSub:
		supressHintSubArg = " -n"
	else:
		supressHintSubArg = ""

	if options.allowDecimalCoords:
		decimalArg = " -d"
	else:
		decimalArg = ""

	dotCount = 0
	seenGlyphCount = 0
	processedGlyphCount = 0
	for name in glyphList:
		prevACIdentifier = None
		seenGlyphCount +=1

		# 	Convert to bez format
		bezString, width, hasHints = fontData.convertToBez(name, removeHints, options.verbose, options.hintAll)
		processedGlyphCount += 1
		if bezString == None:
			continue

		if "mt" not in bezString:
			# skip empty glyphs.
			continue
		# get new fontinfo string if FD array index has changed, as
		# as each FontDict has different alignment zones.
		gid = fontData.getGlyphID(name)
		if isCID: #
			fdIndex = fontData.getfdIndex(gid)
			if not fdIndex == lastFDIndex:
				lastFDIndex = fdIndex
				fdDict = fontData.getFontInfo(psName, path, options.allow_no_blues, options.noFlex, options.vCounterGlyphs, options.hCounterGlyphs, fdIndex)
				with open(tempFI, "w") as fp:
					fp.write(tounicode(fdDict.getFontInfo()))
		else:
			if (fdGlyphDict != None):
				try:
					fdIndex = fdGlyphDict[name][0]
				except KeyError:
					# use default dict.
					fdIndex = 0
				if lastFDIndex != fdIndex:
					lastFDIndex = fdIndex
					fdDict = fontDictList[fdIndex]
					with open(tempFI, "w") as fp:
						fp.write(tounicode(fdDict.getFontInfo()))


		# 	Build autohint point list identifier

		oldBezString = ""
		oldHintBezString = ""
		if (not options.logOnly) and options.usePlistFile:
			# If the glyph is not in the  plist file, then we skip it unless kReHintUnknown is set.
			# If the glyph is in the plist file and the outline has changed, we hint it.
			ACidentifier = makeACIdentifier(bezString)
			try:
				(prevACIdentifier, ACtime, oldBezString, oldHintBezString) =  fontPlist[kACIDKey][name]
			except ValueError:
				(prevACIdentifier, ACtime) =  fontPlist[kACIDKey][name]
				oldBezString = oldHintBezString = ""
			except KeyError:
				pListChanged = 1 # Didn't have an entry in tempList file, so we will add one.
				if hasHints and not options.rehint:
					# Glyphs is hinted, but not referenced in the plist file. Skip it unless options.rehint is seen
					if  not isNewPlistFile:
						# Comment only if there is a plist file; otherwise, we'd be complaining for almost every glyph.
						logMsg("%s Skipping glyph - it has hints, but it is not in the hint info plist file." % aliasName(name))
						dotCount = 0
					continue
			if prevACIdentifier and (prevACIdentifier == ACidentifier): # there is an entry in the plist file and it matches what's in the font.
				if hasHints and not (options.hintAll or options.rehint):
					continue
			else:
				pListChanged = 1

		if options.verbose:
			if fdGlyphDict:
				logMsg("Hinting %s with fdDict %s." % (aliasName(name), fdDict.DictName) )
			else:
				logMsg("Hinting %s." % aliasName(name))
		else:
			logMsg(".,")

		# 	Call auto-hint library on bez string.
		with open(tempBez, "wt") as bp:
			bp.write(tounicode(bezString))

		#print "oldBezString", oldBezString
		#print ""
		#print "bezString", bezString

		if oldBezString != "" and oldBezString == bezString:
			newBezString = oldHintBezString
		else:
			if os.path.exists(tempBezNew):
				os.remove(tempBezNew)
			command = '"%s" %s%s%s%s -s %s -f "%s" "%s"' % (AUTOHINTEXE, verboseArg, suppressEditArg, supressHintSubArg, decimalArg, NEWBEZ_SUFFIX, tempFI, tempBez)
			if  options.debug:
				print(command)
			report = fdkutils.runShellCmd(command)
			if report:
				if not options.verbose:
					logMsg("") # end series of "."
				logMsg(report)
			if os.path.exists(tempBezNew):
				bp = open(tempBezNew, "rt")
				newBezString = bp.read()
				bp.close()
				if options.debug:
					print("Wrote AC fontinfo data file to %s" % tempFI)
					print("Wrote AC output bez file to %s" % tempBezNew)
				else:
					os.remove(tempBezNew)
			else:
				newBezString = None

		if not newBezString:
			if not options.verbose:
				logMsg("")
			logMsg("%s Error - failure in processing outline data." % aliasName(name))
			continue

		if not (("ry" in newBezString[:200]) or ("rb" in newBezString[:200]) or ("rm" in newBezString[:200]) or ("rv" in newBezString[:200])):
			print("No hints added!")

		if options.logOnly:
			continue

		# 	Convert bez to charstring, and update CFF.
		anyGlyphChanged = 1
		fontData.updateFromBez(newBezString, name, width, options.verbose)


		if options.usePlistFile:
			bezString = "%% %s%s%s" % (name, os.linesep, newBezString)
			ACidentifier = makeACIdentifier(bezString)
			# add glyph hint entry to plist file
			if options.allowChanges:
				if prevACIdentifier and (prevACIdentifier != ACidentifier):
					logMsg("\t%s Glyph outline changed" % aliasName(name))
					dotCount = 0

			fontPlist[kACIDKey][name] = (ACidentifier, time.asctime(), bezString, newBezString )

	if not options.verbose:
		print() # print final new line after progress dots.

	if  options.debug:
		print("Wrote input AC bez file to %s" % tempBez)

	if not options.logOnly:
		if anyGlyphChanged:
			logMsg("Saving font file with new hints..." + time.asctime())
			fontData.saveChanges()
		else:
			fontData.close()
			if options.usePlistFile:
				if options.rehint:
					logMsg("No new hints. All glyphs had hints that matched the hint record file %s." % (fontPlistFilePath))
				if options.hintAll:
					logMsg("No new hints. All glyphs had hints that matched the hint history file %s, or were not in the history file and already had hints." % (fontPlistFilePath))
				else:
					logMsg("No new hints. All glyphs were already hinted.")
			else:
				logMsg("No glyphs were hinted.")
	if options.usePlistFile and (anyGlyphChanged or pListChanged):
		#  save font plist file.
		fontPlist.write(fontPlistFilePath)
	if processedGlyphCount != seenGlyphCount:
		logMsg("Skipped %s of %s glyphs." % (seenGlyphCount - processedGlyphCount, seenGlyphCount))
	logMsg("Done with font %s. End time: %s." % (path, time.asctime()))

def main():

	try:
		CheckEnvironment()
	except FDKEnvironmentError as e:
		logMsg(e)
		return

	try:
		options = getOptions()
	except ACOptionParseError as e:
		logMsg(e)
		return

	# verify that all files exist.
	try:
		hintFile(options)
	except (ACFontError, ufotools.UFOParseError) as e:
		logMsg("\t%s" % e)
	if gLogFile:
		gLogFile.close()

	return


if __name__=='__main__':
	main()





from __future__ import print_function, division, absolute_import

import os
import pytest
from shutil import copy2, rmtree

from fontTools.misc.py23 import tobytes

from runner import main as runner
from differ import main as differ
from test_utils import get_input_path, get_expected_path, generate_ttx_dump

TOOL = 'otc2otf'
CMD = ['-t', TOOL]

REGULAR = 'SourceSansPro-Regular.otf'
ITALIC = 'SourceSansPro-It.otf'
BOLD = 'SourceSansPro-Bold.otf'

DATA_DIR = os.path.join(os.path.split(__file__)[0], TOOL + '_data')
TEMP_DIR = os.path.join(DATA_DIR, 'temp_output')


def setup_module():
    """
    Create the temporary output directory
    """
    os.mkdir(TEMP_DIR)


def teardown_module():
    """
    teardown the temporary output directory
    """
    rmtree(TEMP_DIR)


# -----
# Tests
# -----

@pytest.mark.parametrize('ttc_filename, otf_filenames, diff_index', [
    ('RgBd.ttc', [REGULAR, BOLD], 0),
    ('RgIt.ttc', [REGULAR, ITALIC], 1),
    ('RgItBd.ttc', [REGULAR, ITALIC, BOLD], 2),
])
def test_convert(ttc_filename, otf_filenames, diff_index):
    # 'otc2otf' doesn't have an output option. The test makes a temporary
    # directory and copies the input TTC font there. The output OTFs will
    # be saved to that directory as well.
    input_ttc_path = get_input_path(ttc_filename)
    temp_ttc_path = os.path.join(TEMP_DIR, ttc_filename)
    copy2(input_ttc_path, temp_ttc_path)

    fonts_msg = os.linesep.join(
        ['Output font: {}'.format(fname) for fname in otf_filenames])

    stdout_path = runner(CMD + ['-s', '-a', '-f', temp_ttc_path])

    with open(stdout_path, 'rb') as f:
        output = f.read()
    assert tobytes(fonts_msg) in output

    # diff only one of the OTFs
    otf_filename = otf_filenames[diff_index]
    actual_path = os.path.join(TEMP_DIR, otf_filename)
    assert os.path.isfile(actual_path)
    actual_ttx = generate_ttx_dump(actual_path)
    expected_ttx = get_expected_path(otf_filename.replace('.otf', '.ttx'))
    assert differ([expected_ttx, actual_ttx, '-s', '<ttFont sfntVersion'])



from __future__ import print_function, division, absolute_import

import pytest

from runner import main as runner
from differ import main as differ, SPLIT_MARKER
from test_utils import (get_input_path, get_expected_path, get_temp_file_path,
                        generate_ttx_dump)

TOOL = 'otf2ttf'
CMD = ['-t', TOOL]


# -----
# Tests
# -----

@pytest.mark.parametrize('filename', ['sans', 'serif', 'latincid', 'kanjicid'])
def test_convert(filename):
    input_path = get_input_path('{}.otf'.format(filename))
    actual_path = get_temp_file_path()
    runner(CMD + ['-o', 'o', '_{}'.format(actual_path),
                  '-f', input_path])
    actual_ttx = generate_ttx_dump(actual_path)
    expected_ttx = get_expected_path('{}.ttx'.format(filename))
    assert differ([expected_ttx, actual_ttx,
                   '-s',
                   '<ttFont sfntVersion' + SPLIT_MARKER +
                   '    <checkSumAdjustment value=' + SPLIT_MARKER +
                   '    <checkSumAdjustment value=' + SPLIT_MARKER +
                   '    <created value=' + SPLIT_MARKER +
                   '    <modified value='])



# Copyright 2017 Adobe. All rights reserved.

from __future__ import print_function, division, absolute_import

__usage__ = """
buildcff2vf.py  1.14.2 Feb 1 2019
Build a variable font from a designspace file and the UFO master source fonts.

python buildcff2vf.py -h
python buildcff2vf.py -u
python buildcff2vf.py <path to designspace file> (<optional path to output
                                                               variable font>)
"""

__help__ = __usage__ + """
Options:
-p   Use 'post' table format 3.

The script makes a number of assumptions.
1) all the master source fonts are blend compatible in all their data.
2) The source OTF files are in the same directory as the master source
   fonts, and have the same file name but with an extension of '.otf'
   rather than '.ufo'.
3) The master source OTF fonts were built with the companion script
   'buildmasterotfs.py'. This does a first pass of compatibilization
   by using 'tx' with the '-no_opt' option to undo T2 charstring
   optimization applied by makeotf.

The variable font inherits all the OpenType Tables except CFF2 and GPOS
from the default master source font. The default font is flagged in the
designspace file by having the element "<info copy="1" />" in the <source>
element.

The width values and the GPOS positioning data are drawn from all the
master source fonts, so each must be built with with a full set of GPOS
features.

The companion script buildmasterotfs.py will build the master source OTFs
from the designspace file.

Any python interpreter may be used to run the script, as long as it has
installed the latest version of the fonttools module from
https://github.com/fonttools/fonttools
"""

import collections
import io
import logging
import os
import sys

from fontTools import varLib, version as fontToolsVersion
from fontTools.misc.py23 import tobytes
from fontTools.ttLib import TTFont, newTable
from fontTools.cffLib import (VarStoreData, buildOpcodeDict,
                              privateDictOperators)
from fontTools.misc.psCharStrings import T2OutlineExtractor, T2CharString
from fontTools.pens.t2CharStringPen import T2CharStringPen
from fontTools.ttLib.tables import otTables
try:
    import xml.etree.cElementTree as ET
except ImportError:
    import xml.etree.ElementTree as ET
from pkg_resources import parse_version

XML = ET.XML
XMLElement = ET.Element
xmlToString = ET.tostring

kMaxStack = 48
kTempCFFSuffix = ".temp.cff"
kTempCFF2File = "test.cff2"

# setup basic logging to enable fontTools errors to peek thru
logging.basicConfig()


class ACFontError(KeyError):
    pass


class AxisMapError(KeyError):
    pass


def openOpenTypeFile(path):
    try:
        ttFont = TTFont(path)
    except IOError:
        import traceback
        print("\t%s" % (
            traceback.format_exception_only(sys.exc_type, sys.exc_value)[-1]))
        print("Attempted to read font %s as CFF." % path)
        raise ACFontError("Error parsing font file <%s>." % path)

    fontData = CFF2FontData(ttFont, path)
    return fontData


class CFF2FontData(object):
    def __init__(self, ttFont, srcPath):
        self.ttFont = ttFont
        self.srcPath = srcPath
        try:
            self.cffTable = ttFont["CFF "]
            topDict = self.cffTable.cff.topDictIndex[0]
        except KeyError:
            raise ACFontError(
                "Error: font is not a CFF font <%s>." % srcPath)

        # for identifier in glyph-list:
        # Get charstring.
        self.topDict = topDict
        self.isCID = hasattr(topDict, "FDArray")
        self.charStrings = topDict.CharStrings
        self.charStringIndex = self.charStrings.charStringsIndex
        self.allowDecimalCoords = False


class OpListExtractor(T2OutlineExtractor):
    def __init__(self, pen, subrs, globalSubrs, nominalWidthX, defaultWidthX):
        super(OpListExtractor, self).__init__(pen, subrs, globalSubrs,
                                              nominalWidthX, defaultWidthX)
        self.seenHints = None
        self.hintmaskLen = 0
        self.hints = []

    def countHints(self):
        args = self.popallWidth()
        self.hintCount = self.hintCount + len(args) // 2
        return args

    def op_hstem(self, index):
        self.seenHints = 'h'
        args = self.countHints()
        self.pen.hstem(args)

    def op_vstem(self, index):
        self.seenHints = 'v'
        args = self.countHints()
        self.pen.vstem(args)

    def op_hstemhm(self, index):
        self.seenHints = 'h'
        args = self.countHints()
        self.pen.hstemhm(args)

    def op_vstemhm(self, index):
        self.seenHints = 'v'
        args = self.countHints()
        self.pen.vstemhm(args)

    def op_hintmask(self, index):
        if not self.sawMoveTo:
            # are processing the first hintmask.
            if not self.seenHints:
                args = self.countHints()
                if args:
                    self.pen.op_hstemhm(args)
            elif self.seenHints == 'h':
                args = self.countHints()
                if args:
                    # There may be hstem hints but no vstem hints.
                    self.pen.vstemhm(args)
        if not self.hintmaskLen:
            self.countHints()
            self.hintmaskLen = (self.hintCount + 7) // 8
        hintMaskBytes, index = self.callingStack[-1].getBytes(
            index, self.hintmaskLen)
        self.pen.hintmask(hintMaskBytes)
        return hintMaskBytes, index

    def op_cntrmask(self, index):
        if not self.sawMoveTo:
            # are processing the first hintmask.
            if not self.seenHints:
                args = self.countHints()
                if args:
                    self.pen.op_hstemhm(args)
            elif self.seenHints == 'h':
                args = self.countHints()
                if args:
                    # There may be hstem hints but no vstem hints.
                    self.pen.vstemhm(args)
        if not self.hintmaskLen:
            self.countHints()
            self.hintmaskLen = (self.hintCount + 7) // 8
        hintMaskBytes, index = self.callingStack[-1].getBytes(
            index, self.hintmaskLen)
        self.pen.cntrmask(hintMaskBytes)
        return hintMaskBytes, index


class OpListPen(T2CharStringPen):
    # We use this to de-optimize the T2Charstrings
    # to just rmoveto, rlineto, and rrcurveto.
    def __init__(self, supportHints):
        super(OpListPen, self).__init__(0, {})
        self.opList = []
        self.absMovetoPt = [0, 0]
        self.supportHints = supportHints

    def _moveTo(self, pt):
        self.opList.append(["rmoveto", self._p(pt)])
        self.absMovetoPt = list(self._p0)

    def _lineTo(self, pt):
        self.opList.append(["rlineto", self._p(pt)])

    def _curveToOne(self, pt1, pt2, pt3):
        _p = self._p
        self.opList.append(["rrcurveto", _p(pt1) + _p(pt2) + _p(pt3)])

    def _closePath(self):
        # Add closing lineto if start and end path are not the same.
        # We need this to compatibilize different master designs where
        # the the last path op in one design is a rlineto and is omitted,
        # while the last path op in another is an rrcurveto.
        if (self._p0[0] != self.absMovetoPt[0]) or (
           self._p0[1] != self.absMovetoPt[1]):
            self.opList.append(["rlineto", [self._p0[0] - self.absMovetoPt[0],
                               self._p0[1] - self.absMovetoPt[1]]])

    def vstem(self, args):
        if self.supportHints:
            self.opList.append(["vstem", args])

    def hstem(self, args):
        if self.supportHints:
            self.opList.append(["hstem", args])

    def vstemhm(self, args):
        if self.supportHints:
            self.opList.append(["vstemhm", args])

    def hstemhm(self, args):
        if self.supportHints:
            self.opList.append(["hstemhm", args])

    def hintmask(self, hintMaskBytes):
        if self.supportHints:
            self.opList.append(["hintmask", [hintMaskBytes]])

    def cntrmask(self, hintMaskBytes):
        if self.supportHints:
            self.opList.append(["cntrmask", [hintMaskBytes]])

    def getCharString(self, private=None, globalSubrs=None):
        return self.opList


class CFF2GlyphData(object):
    hintOpList = ('hintmask', 'cntrmask', 'hstem', 'vstem', 'hstemhm',
                  'vstemhm')

    def __init__(self, glyphName, gid, masterFontList):
        self.glyphName = glyphName
        self.gid = gid
        self.masterFontList = masterFontList
        self.charstringList = []
        self.mmCharString = None

    def addCharString(self, t2CharString):
        self.charstringList.append(t2CharString)

    def buildOpList(self, t2Index, supportHints):
        t2String = self.charstringList[t2Index]
        t2Pen = OpListPen(supportHints)
        subrs = getattr(t2String.private, "Subrs", [])
        extractor = OpListExtractor(t2Pen, subrs, t2String.globalSubrs,
                                    t2String.private.nominalWidthX,
                                    t2String.private.defaultWidthX)
        extractor.execute(t2String)
        opList = t2Pen.getCharString(t2String.private, t2String.globalSubrs)
        if t2Index == 0:
            # For the master font path, promote all coordinates to a list.
            for _, ptList in opList:
                for i, item in enumerate(ptList):
                    ptList[i] = [item]
        return opList

    def buildMMData(self):
        # Build MM charstring, and list of  points.
        # First, build a list of path entries for the default master.
        # Each entry is [opName,  [ [arg0], [arg1],..[argn] ].
        # Note that each arg is a list: this is so we can add the corresponding
        # arg from each successive master font design, and end up with
        # [opName,  [ [arg.0.0,...,arg.0.k-1],
        # [arg.1.0,...,arg.1.k-1],..,[arg.n-1.0,...,arg.n-1.k-1] ]
        # where n is the number of arguments for the opName,
        # and k is the number of master designs.
        supportHints = True
        self.opList = opList = self.buildOpList(0, supportHints)
        numOps = len(opList)
        # For each other master in turn,
        #  add the args for each opName to pointList
        # deal with incompatible path lists because:
        # converting from UFO to Type1 can convert flat curves to lines.
        blendError = False
        i = 1
        while i < len(self.charstringList):
            opList2 = self.buildOpList(i, supportHints)
            i += 1
            opIndex = op2Index = 0
            while opIndex < numOps:
                masterOp, masterPointList = opList[opIndex]
                if (not supportHints) and masterOp in self.hintOpList:
                    opIndex += 1
                    continue

                # Do minor work to compatibilize charstrings.
                try:
                    token, pointList = opList2[op2Index]
                except IndexError:
                    print("Path mismatch: different number of points. "
                          "Glyph name: %s. " % self.glyphName,
                          "font index: %s. " % (i - 1),
                          "master font path: %s. " %
                          self.masterFontList[0].srcPath,
                          "Delta font path: %s." %
                          self.masterFontList[i - 1].srcPath)
                    blendError = True
                    break

                if (not supportHints) and token in self.hintOpList:
                    op2Index += 1
                    continue

                if token == 'rlineto' and masterOp == 'rrcurveto':
                    self.updateLineToCurve(pointList)
                elif masterOp == 'rlineto' and token == 'rrcurveto':
                    self.updateLineToCurve(masterPointList)
                    opList[opIndex] = ['rrcurveto', masterPointList]
                elif token != masterOp:
                    if supportHints and (masterOp in self.hintOpList or (
                       token in self.hintOpList)):
                        # restart with hint support off.
                        supportHints = False
                        self.opList = opList = self.buildOpList(
                            0, supportHints)
                        numOps = len(opList)
                        i = 1
                        break
                    print("Path mismatch in glyph %s, master index %s. %s." % (
                          self.glyphName, i - 1,
                          self.masterFontList[i - 1].srcPath))
                    blendError = True
                    break

                try:
                    self.mergePointList(masterPointList, pointList,
                                        masterOp in ('hintmask', 'cntrmask'))
                except IndexError:
                    if supportHints and masterOp in self.hintOpList or (
                       token in self.hintOpList):
                        # restart with hint support off.
                        supportHints = False
                        self.opList = opList = self.buildOpList(
                            0, supportHints)
                        numOps = len(opList)
                        i = 1
                        break
                    else:
                        print("Path mismatch 2", self.glyphName, i,
                              self.masterFontList[i - 1].srcPath)
                        blendError = True
                        break
                opIndex += 1
                op2Index += 1
        # Now pointList is the list of number lists' first master value,
        # folowed by the other master values.
        if not supportHints:
            print("\t skipped incompatible hint data for", self.glyphName)
        return blendError

    def mergePointList(self, masterPointList, pointList, isHintMask):
        i = 0
        while i < len(masterPointList):
            if isHintMask:
                # can't blend hint masks. Just use the first one.
                masterPointList[i].append(masterPointList[i][0])
            else:
                masterPointList[i].append(pointList[i])
            i += 1

    def updateLineToCurve(self, pointList):
        # pointList has a lineto that needs to be a flat curve.
        # Convert to curve by pre-pending [0,0], and appending [0,0]
        if isinstance(pointList[0], list):
            numMasters = len(pointList[0])
            pointList.insert(0, [0] * numMasters)
            pointList.insert(0, [0] * numMasters)
            pointList.append([0] * numMasters)
            pointList.append([0] * numMasters)
        else:
            pointList.insert(0, 0)
            pointList.insert(0, 0)
            pointList.append(0)
            pointList.append(0)


class AxisValueRecord(object):
    def __init__(self, nameID, axisValue, flagValue, valueIndex, axisIndex):
        self.nameID = nameID
        self.axisValue = axisValue
        self.flagValue = flagValue
        self.valueIndex = valueIndex
        self.axisIndex = axisIndex


def getInsertGID(origGID, fontGlyphNameList, mergedGlyphNameList):
    if origGID == 0:
        return origGID
    newGID = 0
    # try stepping down in GID.
    gid = origGID - 1
    while gid >= 0:
        prevName = fontGlyphNameList[gid]
        newGID = mergedGlyphNameList.index(prevName)
        if newGID >= 0:
            newGID += 1
            return newGID
        gid -= 1

    # try stepping up in GID.
    numGlyphs = len(fontGlyphNameList)
    gid = origGID + 1
    while gid < numGlyphs:
        prevName = fontGlyphNameList[gid]
        newGID = mergedGlyphNameList.index(prevName)
        if newGID >= 0:
            newGID -= 1
            return newGID
        gid += 1

    return None


def buildMasterList(inputPaths):
    blendError = False
    cff2FontList = []
    # Collect all the charstrings.
    cff2GlyphList = collections.OrderedDict()
    print("Opening default font", inputPaths[0])
    baseFont = openOpenTypeFile(inputPaths[0])
    cff2FontList.append(baseFont)

    # Get the master source data for all the glyphs.
    fontGlyphList = baseFont.ttFont.getGlyphOrder()
    for glyphName in fontGlyphList:
        gid = baseFont.charStrings.charStrings[glyphName]
        cff2GlyphData = CFF2GlyphData(glyphName, gid, cff2FontList)
        cff2GlyphList[glyphName] = cff2GlyphData

        t2CharString = baseFont.charStringIndex[gid]
        cff2GlyphData.addCharString(t2CharString)

    for fontPath in inputPaths[1:]:
        print("Opening", fontPath)
        masterFont = openOpenTypeFile(fontPath)
        cff2FontList.append(masterFont)
        for glyphName in fontGlyphList:
            cff2GlyphData = cff2GlyphList[glyphName]
            gid = masterFont.charStrings.charStrings[glyphName]
            if gid != cff2GlyphData.gid:
                raise ACFontError("GID in master font did not match GID in "
                                  "base font: %s." % glyphName)
            t2CharString = masterFont.charStringIndex[gid]
            cff2GlyphData.addCharString(t2CharString)

    # Now build MM versions.
    print("Reading glyph data...")
    blendError = False
    for glyphName in fontGlyphList:
        cff2GlyphData = cff2GlyphList[glyphName]
        blendError = cff2GlyphData.buildMMData()
        if blendError:
            break
    if blendError:
        print("Failed to blend master designs.")
        return None, None, None, None, blendError

    # Now get the MM data for the Private Dict.
    # Not all fonts will have the same keys.
    # The first time a key is encountered, we fill out the
    # value list for that key for all fonts, using the values from
    # the first font which has the key. When other fonts with the
    # key are encountered, the real values will overwrite the default values.
    privateDictList = []
    numMasters = len(cff2FontList)
    curFont = cff2FontList[0]
    isCID = curFont.isCID
    if isCID:
        numFDArray = len(cff2FontList[0].topDict.FDArray)
    else:
        numFDArray = 1
    for fdIndex in range(numFDArray):
        blendedPD = {}
        privateDictList.append(blendedPD)
        for master_index in range(numMasters):
            curFont = cff2FontList[master_index]
            if isCID:
                pd = curFont.topDict.FDArray[fdIndex].Private
            else:
                pd = curFont.topDict.Private
            for key, value in pd.rawDict.items():
                try:
                    valList = blendedPD[key]
                    if isinstance(valList[0], list):
                        lenValueList = len(value)
                        if lenValueList != len(valList):
                            print("Error: the number of items in the",
                                  "value list for key", key,
                                  "in the Private dict of FDDict index",
                                  fdIndex, "in master font", master_index,
                                  "is different than for previous masters.")
                            print("Failed to blend master designs.")
                            blendError = True
                            return None, None, None, None, blendError
                        for j in range(lenValueList):
                            val = value[j]
                            valList[j][master_index] = val
                    else:
                        blendedPD[key][master_index] = value
                except KeyError:
                    if isinstance(value, list):
                        lenValueList = len(value)
                        valList = [None] * lenValueList
                        for j in range(lenValueList):
                            vList = [value[j]] * numMasters
                            valList[j] = vList
                        blendedPD[key] = valList
                    else:
                        blendedPD[key] = [value] * numMasters

    return baseFont, privateDictList, cff2GlyphList, fontGlyphList, blendError


def pointsDiffer(pointList):
    val = False
    p0 = pointList[0]
    for p in pointList[1:]:
        if p != p0:
            val = True
            break
    return val


def appendBlendOp(op, pointList, varModel):
    blendStack = []
    numBlends = len(pointList)
    # pointList is arranged as:
    # [
    #   [m0, m1..mn] # values for each master for arg 0
    #   [m0, m1..mn] # values for each master for arg 1
    #   ...
    #   [m0, m1..mn] # values for each master for arg numBlends
    # ]
    # The CFF2 blend operator expects first the series of args 0-numBlends
    # from the first master
    blendList = []
    if op in ('hintmask', 'cntrmask'):
        blendStack.append(op)
        blendStack.append(pointList[0][0])
        return blendStack
    for argEntry in pointList:
        masterArg = argEntry[0]
        if pointsDiffer(argEntry):
            blendStack.append(masterArg)
            blendList.append(argEntry)
        else:
            if blendList:
                for masterValues in blendList:
                    deltas = varModel.getDeltas(masterValues)
                    # First item in 'deltas' is the default master value;
                    # for CFF2 data, that has already been written.
                    blendStack.extend(deltas[1:])
                numBlends = len(blendList)
                blendStack.append(numBlends)
                blendStack.append('blend')
                blendList = []
            blendStack.append(masterArg)
    if blendList:
        for masterValues in blendList:
            deltas = varModel.getDeltas(masterValues)
            # First item in 'deltas' is the default master value;
            # for CFF2 data, that has already been written.
            blendStack.extend(deltas[1:])
        numBlends = len(blendList)
        blendStack.append(numBlends)
        blendStack.append('blend')
        blendList = []
    blendStack.append(op)
    return blendStack


def buildMMCFFTables(baseFont, privateDictList, cff2GlyphList, varModel):

    # Build the blended PrivateDicts.
    opCodeDict = buildOpcodeDict(privateDictOperators)
    blendedPD = privateDictList[0]
    if baseFont.isCID:
        numFDArray = len(baseFont.topDict.FDArray)
    else:
        numFDArray = 1
    for fdIndex in range(numFDArray):
        if baseFont.isCID:
            pd = baseFont.topDict.FDArray[fdIndex]
            blendedPD = privateDictList[fdIndex]
        else:
            pd = baseFont.topDict.Private
            blendedPD = privateDictList[0]

        for key in blendedPD.keys():
            argType = opCodeDict[key][1]
            valList = blendedPD[key]
            if argType == 'delta':
                # If all masters are the same for each entry in the list, then
                # save a non-blend version; else save the blend version
                needsBlend = False
                for blendList in valList:
                    if pointsDiffer(blendList):
                        needsBlend = True
                        break
                dataList = []
                if needsBlend:
                    prevBlendList = [0] * len(valList[0])
                    for blendList in valList:
                        # convert blend list from absolute values to relative
                        # values from the previous blend list.
                        relBlendList = [(val - prevBlendList[i]) for i, val in
                                        enumerate(blendList)]
                        prevBlendList = blendList
                        deltas = varModel.getDeltas(relBlendList)
                        # For PrivateDict BlueValues, the default font
                        # values are absolute, not relative.
                        deltas[0] = blendList[0]
                        dataList.append(deltas)
                else:
                    for blendList in valList:
                        dataList.append(blendList[0])
            else:
                if pointsDiffer(valList):
                    dataList = varModel.getDeltas(valList)
                else:
                    dataList = valList[0]

            pd.rawDict[key] = dataList

    # Now update all the charstrings.
    fontGlyphList = baseFont.ttFont.getGlyphOrder()
    for glyphName in fontGlyphList:
        gid = baseFont.charStrings.charStrings[glyphName]
        t2CharString = baseFont.charStringIndex[gid]
        cff2GlyphData = cff2GlyphList[glyphName]
        t2CharString.decompile()
        newProgram = []
        for op, pointList in cff2GlyphData.opList:
            blendStack = appendBlendOp(op, pointList, varModel)
            newProgram.extend(blendStack)
        t2CharString = T2CharString(private=t2CharString.private,
                                    globalSubrs=t2CharString.globalSubrs)
        baseFont.charStringIndex[gid] = t2CharString
        t2CharString.program = newProgram
        t2CharString.private.defaultWidthX = 0
        t2CharString.private.nominalWidthX = 0
        t2CharString.compile(isCFF2=True)


def addCFFVarStore(baseFont, varModel, varFont):
    supports = varModel.supports[1:]
    fvarTable = varFont['fvar']
    axisKeys = [axis.axisTag for axis in fvarTable.axes]
    varTupleList = varLib.builder.buildVarRegionList(supports, axisKeys)
    varTupleIndexes = list(range(len(supports)))
    varDeltasCFFV = varLib.builder.buildVarData(varTupleIndexes, None, False)
    varStoreCFFV = varLib.builder.buildVarStore(varTupleList, [varDeltasCFFV])

    cffTable = baseFont.cffTable
    topDict = cffTable.cff.topDictIndex[0]
    topDict.VarStore = VarStoreData(otVarStore=varStoreCFFV)


def addNamesToPost(ttFont, fontGlyphList):
    postTable = ttFont['post']
    postTable.glyphOrder = ttFont.glyphOrder = fontGlyphList
    postTable.formatType = 2.0
    postTable.extraNames = []
    postTable.mapping = {}
    postTable.compile(ttFont)


def convertCFFtoCFF2(baseFont, masterFont, post_format_3=False):
    # base font contains the CFF2 blend data, but with the table tag 'CFF '
    # all the CFF fields. Remove all the fields that were removed in the
    # CFF2 spec.

    cffTable = baseFont.ttFont['CFF ']
    del masterFont['CFF ']
    cffTable.cff.convertCFFToCFF2(masterFont)
    newCFF2 = newTable("CFF2")
    newCFF2.cff = cffTable.cff
    newCFF2.compile(masterFont)
    masterFont['CFF2'] = newCFF2
    if post_format_3:
        masterFont['post'].formatType = 3.0


def reorderMasters(mapping, masterPaths):
    numMasters = len(masterPaths)
    new_order = [0] * numMasters
    for i in range(numMasters):
        new_order[mapping[i]] = masterPaths[i]
    return new_order


def fixVerticalMetrics(masterFont, masterPaths):
    # Fix the head table bbox, and OS/2 winAscender/Descender.
    # Record the largest BBOX extents, and use these values.
    print("Updating vertical metrics to most extreme in source fonts.")
    print("\tname old new")

    xMax, yMax, xMin, yMin = ([] for i in range(4))
    ascender, descender = ([] for i in range(2))
    gap = 0

    for fPath in masterPaths:
        srcFont = TTFont(fPath)

        headTable = srcFont['head']
        xMax.append(headTable.xMax)
        yMax.append(headTable.yMax)
        xMin.append(headTable.xMin)
        yMin.append(headTable.yMin)

        os2Table = srcFont['OS/2']
        ascender.append(os2Table.sTypoAscender)
        descender.append(os2Table.sTypoDescender)

    xMax = max(xMax)
    yMax = max(yMax)
    xMin = min(xMin)
    yMin = min(yMin)
    ascender = max(ascender)
    descender = min(descender)

    headTable = masterFont['head']
    print("\theadTable.xMax", headTable.xMax, xMax)
    print("\theadTable.yMax", headTable.yMax, yMax)
    print("\theadTable.xMin", headTable.xMin, xMin)
    print("\theadTable.yMin", headTable.yMin, yMin)
    headTable.xMax = xMax
    headTable.yMax = yMax
    headTable.xMin = xMin
    headTable.yMin = yMin

    # IMPORTANT: The vertical metrics are deliberately set in a
    # way that ensures cross-platform compatibility in browsers.
    # This approach to setting vertical metrics is different from
    # what Adobe Type has used for many years in their fonts.

    hheaTable = masterFont['hhea']
    print("\thheaTable.ascent", hheaTable.ascent, yMax)
    print("\thheaTable.descent", hheaTable.descent, yMin)
    print("\thheaTable.lineGap", hheaTable.lineGap, gap)
    hheaTable.ascent = yMax
    hheaTable.descent = yMin
    hheaTable.lineGap = gap

    OS2Table = masterFont['OS/2']
    print("\tOS2Table.usWinAscent", OS2Table.usWinAscent, yMax)
    print("\tOS2Table.usWinDescent", OS2Table.usWinDescent, abs(yMin))
    OS2Table.usWinAscent = yMax
    OS2Table.usWinDescent = abs(yMin)
    print("\tOS2Table.sTypoAscender", OS2Table.sTypoAscender, ascender)
    print("\tOS2Table.sTypoDescender", OS2Table.sTypoDescender, descender)
    print("\tOS2Table.sTypoLineGap", OS2Table.sTypoLineGap, gap)
    OS2Table.sTypoAscender = ascender
    OS2Table.sTypoDescender = descender
    OS2Table.sTypoLineGap = gap


def getNewAxisValue(seenCoordinate, fvarInstance, axisTag):
    # Return None if:
    #   have already seen this axis coordinate
    #   any of the other axis coordinates are non-zero
    axisValue = fvarInstance.coordinates[axisTag]
    if axisValue in seenCoordinate:
        return None
    for tag, value in fvarInstance.coordinates.items():
        if tag == axisTag:
            continue
        if value != 0:
            return
    seenCoordinate[axisValue] = axisTag
    return axisValue, fvarInstance.subfamilyNameID


def addAxisValueData(xmlData, prevEntry, axisEntry, nextEntry, linkDelta):
    if linkDelta is not None:
        _format = 3
    elif (prevEntry is None) and (nextEntry is None):
        _format = 1
    else:
        _format = 2

    xmlData.append("\t\t<AxisValue index=\"%s\" Format=\"%s\">" % (
        axisEntry.valueIndex, _format))
    xmlData.append("\t\t\t<AxisIndex value=\"%s\" />" % (axisEntry.axisIndex))
    xmlData.append("\t\t\t<Flags value=\"%s\" />" % (axisEntry.flagValue))
    xmlData.append("\t\t\t<ValueNameID value=\"%s\" />" % (axisEntry.nameID))
    if _format == 1:
        xmlData.append("\t\t\t<Value value=\"%s\" />" % (axisEntry.axisValue))
    elif _format == 2:
        if prevEntry is None:
            minValue = axisEntry.axisValue
        else:
            minValue = (axisEntry.axisValue + prevEntry.axisValue) / 2.0
        if nextEntry is None:
            maxValue = axisEntry.axisValue
        else:
            maxValue = (axisEntry.axisValue + nextEntry.axisValue) / 2.0
        xmlData.append("\t\t\t<NominalValue value=\"%s\" />" % (
            axisEntry.axisValue))
        xmlData.append("\t\t\t<RangeMinValue value=\"%s\" />" % (minValue))
        xmlData.append("\t\t\t<RangeMaxValue value=\"%s\" />" % (maxValue))
    elif _format == 3:
        xmlData.append("\t\t\t<Value value=\"%s\" />" % (axisEntry.axisValue))
        xmlData.append("\t\t\t<LinkedValue value=\"%s\" />" % (linkDelta))
    xmlData.append("\t\t</AxisValue>")


def makeSTAT(fvar):
    newSTAT = newTable("STAT")
    newSTAT.table = otTables.STAT()
    xmlData = ["<?xml version=\"1.0\" encoding=\"UTF-8\"?>",
               "<ttFont sfntVersion=\"OTTO\" ttLibVersion=\"3.11\">",
               "<STAT>",
               "\t<Version value=\"0x00010001\"/>",
               "\t<DesignAxisRecordSize value=\"8\"/>",
               "\t<DesignAxisRecord>"]
    for i, a in enumerate(fvar.axes):
        xmlData.append("\t\t<Axis index=\"%s\">" % (i))
        xmlData.append("\t\t\t<AxisTag value=\"%s\"/>" % (a.axisTag))
        xmlData.append("\t\t\t<AxisNameID value=\"%s\"/>" % (a.axisNameID))
        xmlData.append("\t\t\t<AxisOrdering value=\"%s\"/>" % (i))
        xmlData.append("\t\t</Axis>")

    xmlData.append("\t</DesignAxisRecord>")
    xmlData.append("\t<AxisValueArray>")
    # For each axis in turn, cycle through the named instances and
    # make a AxisValue entry for the instance of each new named instance when
    # the other axes are at 0.
    numAxes = len(fvar.axes)
    numInstances = len(fvar.instances)
    fallBackNameID = None
    for axisIndex in range(numAxes):
        axisTag = fvar.axes[axisIndex].axisTag
        seenCoordinate = {}
        axisList = []
        valueIndex = 0
        for j in range(numInstances):
            flagValue = 0
            instance = fvar.instances[j]
            entry = getNewAxisValue(seenCoordinate, instance, axisTag)
            if entry is None:
                continue
            axisValue, nameID = entry
#           if (axisValue == 0) and (axisTag == 'wght'):
#               # the style name from any other axis will probably
#               # already be elided from the style name.
#               fallBackNameID = nameID
#               flagValue = 2
            axisEntry = AxisValueRecord(nameID, axisValue, flagValue,
                                        valueIndex, axisIndex)
            axisList.append(axisEntry)
            valueIndex += 1

        numEntries = len(axisList)
        linkDelta = None
        if numEntries == 0:
            continue
        if numEntries == 1:
            addAxisValueData(xmlData, None, axisList[0], None, linkDelta)
            continue
        if numEntries == 2:
            addAxisValueData(xmlData, None, axisList[0], axisList[1],
                             linkDelta)
            addAxisValueData(xmlData, axisList[0], axisList[1], None,
                             linkDelta)
            continue

        prevEntry = nextEntry = None
        for j in range(numEntries - 1):
            axisEntry = axisList[j]
            nextEntry = axisList[j + 1]
            addAxisValueData(xmlData, prevEntry, axisEntry, nextEntry,
                             linkDelta)
            prevEntry = axisEntry
        addAxisValueData(xmlData, axisEntry, nextEntry, None, linkDelta)

    xmlData.append("\t</AxisValueArray>")

    if fallBackNameID is None:
        fallBackNameID = 2
    xmlData.append("\t<ElidedFallbackNameID value=\"%s\" />" % (
        fallBackNameID))
    xmlData.append("</STAT>")
    xmlData.append("</ttFont>")
    xmlData.append("")
    xmlData = os.linesep.join(xmlData)
    return xmlData


def addSTATTable(varFont, varFontPath):
    print("Adding STAT table")
    kSTAT_OverrideName = "override.STAT.ttx"
    statPath = os.path.dirname(varFontPath)
    statPath = os.path.join(statPath, kSTAT_OverrideName)
    if not os.path.exists(statPath):
        print("Note: Generating simple STAT table from 'fvar' table in "
              "'%s'." % (statPath))
        fvar = varFont["fvar"]
        xmlSTATData = tobytes(makeSTAT(fvar))
        statFile = io.BytesIO(xmlSTATData)
        varFont.importXML(statFile)
        varFont.saveXML(statPath, tables=["STAT"])
    else:
        varFont.importXML(statPath)


def buildCFF2Font(varFontPath, varFont, varModel, masterPaths,
                  post_format_3=False):
    """Build CFF2 font from the master designs. default font is first."""
    numMasters = len(masterPaths)
    inputPaths = reorderMasters(varModel.mapping, masterPaths)
    varModel.mapping = varModel.reverseMapping = range(numMasters)
    # Since we have re-ordered the master master data to be the same
    # as the varModel.location order, the mappings are flat.
    (baseFont, privateDictList, cff2GlyphList, fontGlyphList, blendError) = \
        buildMasterList(inputPaths)
    if blendError:
        return blendError
    buildMMCFFTables(baseFont, privateDictList, cff2GlyphList, varModel)
    addCFFVarStore(baseFont, varModel, varFont)
    addNamesToPost(varFont, fontGlyphList)
    convertCFFtoCFF2(baseFont, varFont, post_format_3)
    # fixVerticalMetrics(varFont, masterPaths)
    addSTATTable(varFont, varFontPath)
    varFont.save(varFontPath)
    return blendError


def otfFinder(s):
    return s.replace('.ufo', '.otf')


def run(args=None):
    post_format_3 = False
    if not args:
        args = sys.argv[1:]
    if '-u' in args:
        print(__usage__)
        return
    if '-h' in args:
        print(__help__)
        return
    if '-p' in args:
        post_format_3 = True
        args.remove('-p')

    if parse_version(fontToolsVersion) < parse_version("3.19"):
        print("Quitting. The Python fonttools module must be at least 3.19.0 "
              "in order for buildcff2vf to work.")
        return

    if len(args) == 2:
        designSpacePath, varFontPath = args
    elif len(args) == 1:
        designSpacePath = args[0]
        varFontPath = os.path.splitext(designSpacePath)[0] + '.otf'
    else:
        print(__usage__)
        return

    if os.path.exists(varFontPath):
        os.remove(varFontPath)
    varFont, varModel, masterPaths = varLib.build(designSpacePath,
                                                  otfFinder, exclude=("CFF2",))

    blendError = buildCFF2Font(varFontPath, varFont, varModel, masterPaths,
                               post_format_3)
    if not blendError:
        print("Built variable font '%s'" % (varFontPath))


if __name__ == '__main__':
    run()



from __future__ import print_function, division, absolute_import

import os
import pytest
import subprocess32 as subprocess
import tempfile

from runner import main as runner
from differ import main as differ, SPLIT_MARKER
from test_utils import get_input_path, get_expected_path, get_temp_file_path

TOOL = 'tx'
CMD = ['-t', TOOL]


def _get_temp_dir_path():
    return tempfile.mkdtemp()


def _get_extension(in_format):
    if 'ufo' in in_format:
        return '.ufo'
    elif in_format == 'type1':
        return '.pfa'
    return '.' + in_format


PDF_SKIP = [
    '/Creator' + SPLIT_MARKER +
    '/Producer' + SPLIT_MARKER +
    '/CreationDate' + SPLIT_MARKER +
    '/ModDate' + SPLIT_MARKER +
    '(Date:' + SPLIT_MARKER +
    '(Time:',
]

PDF_SKIP_REGEX = [
    '^.+30.00 Td',
    '^.+0.00 Td',
]

PS_SKIP = [
    '0 740 moveto (Filename:' + SPLIT_MARKER +
    '560 (Date:' + SPLIT_MARKER +
    '560 (Time:'
]


# -----------
# Basic tests
# -----------

@pytest.mark.parametrize('arg', ['-h', '-v', '-u'])
def test_exit_known_option(arg):
    assert subprocess.call([TOOL, arg]) == 0


@pytest.mark.parametrize('arg', ['-z', '-foo'])
def test_exit_unknown_option(arg):
    assert subprocess.call([TOOL, arg]) == 1


# -------------
# Convert tests
# -------------

@pytest.mark.parametrize('to_format', [
    'ufo2',
    'ufo3',
    'type1',
    'svg',
    'mtx',
    'afm',
    'pdf',
    'ps',
    # 'cff',
])
@pytest.mark.parametrize('from_format', [
    'ufo2',
    'ufo3',
    'type1',
])
def test_convert(from_format, to_format):
    from_ext = _get_extension(from_format)
    to_ext = _get_extension(to_format)

    # input filename
    from_filename = from_format + from_ext

    # expected filename
    exp_filename = from_format + to_ext

    # runner args
    if 'ufo' in to_format:
        save_path = os.path.join(_get_temp_dir_path(), 'font.ufo')
    else:
        save_path = get_temp_file_path()

    # diff mode
    if to_format == 'cff':
        diff_mode = ['-m', 'bin']
    else:
        diff_mode = []

    # skip items
    regex_skip = []
    skip = []
    if to_format == 'afm':
        skip = ['Comment Creation Date:' + SPLIT_MARKER + 'Comment Copyright']
    elif to_format == 'pdf':
        skip = PDF_SKIP[:]
        regex_skip = PDF_SKIP_REGEX[:]
    elif to_format == 'ps':
        skip = PS_SKIP[:]
    if skip:
        skip.insert(0, '-s')
    if regex_skip:
        for regex in regex_skip:
            skip.append('-r')
            skip.append(regex)

    # format arg fix
    if to_format in ('ufo2', 'ufo3'):
        format_arg = 'ufo'
    elif to_format == 'type1':
        format_arg = 't1'
    else:
        format_arg = to_format

    runner(CMD + ['-a', '-f', get_input_path(from_filename), save_path,
                  '-o', format_arg])
    expected_path = get_expected_path(exp_filename)
    assert differ([expected_path, save_path] + skip + diff_mode)


# ----------
# Dump tests
# ----------

@pytest.mark.parametrize('args, exp_filename', [
    ([], 'type1.dump1.txt'),
    (['0'], 'type1.dump0.txt'),
    (['dump', '0'], 'type1.dump0.txt'),
    (['1'], 'type1.dump1.txt'),
    (['2'], 'type1.dump2.txt'),
    (['3'], 'type1.dump3.txt'),
    (['4'], 'type1.dump4.txt'),
    (['5'], 'type1.dump5.txt'),
    (['6'], 'type1.dump6.txt'),
    (['6', 'd'], 'type1.dump6d.txt'),
    (['6', 'n'], 'type1.dump6n.txt'),
])
def test_dump(args, exp_filename):
    if args:
        args.insert(0, '-o')

    if any([arg in args for arg in ('4', '5', '6')]):
        skip = []
    else:
        skip = ['-s', '## Filename']

    actual_path = runner(CMD + ['-s', '-f', 'type1.pfa'] + args)
    expected_path = get_expected_path(exp_filename)
    assert differ([expected_path, actual_path] + skip)


# ----------
# CFF2 tests
# ----------

@pytest.mark.parametrize('args, exp_filename', [
    ([], 'SourceCodeVar-Roman_CFF2'),
    (['*S', '*b', 'std'], 'SourceCodeVar-Roman_CFF2_subr'),  # subroutinize
])
def test_cff2_extract(args, exp_filename):
    # read CFF2 VF, write CFF2 table
    font_path = get_input_path('SourceCodeVariable-Roman.otf')
    cff2_path = get_temp_file_path()
    runner(CMD + ['-a', '-f', font_path, cff2_path, '-o', 'cff2'] + args)
    expected_path = get_expected_path(exp_filename)
    assert differ([expected_path, cff2_path, '-m', 'bin'])


def test_cff2_sub_dump():
    # Dump a subroutinized CFF2 font. This is a J font with 64K glyphs,
    # and almost every subr and charstring is a single subr call.
    # A good test for problems with charstrings with no endchar operator.
    actual_path = runner(CMD + ['-s', '-o', 'dump', '6', 'g', '_21847',
                                '-f', 'CFF2-serif-sub.cff2'])
    expected_path = get_expected_path('CFF2-serif-sub.cff2.txt')
    assert differ([expected_path, actual_path])


def test_varread_pr355():
    # read CFF2 VF, write Type1 snapshot
    actual_path = runner(CMD + ['-s', '-o', 't1', '-f', 'cff2_vf.otf'])
    expected_path = get_expected_path('cff2_vf.pfa')
    assert differ([expected_path, actual_path])


def test_cff2_no_vf_bug353():
    # read CFF2 WITHOUT VF info, write a CFF2 out. 'regular_CFF2.otf'
    # is derived by taking the regular.otf file from the sfntdiff
    # 'input_data' directory, and converting the CFF table to CFF2.
    font_path = get_input_path('regular_CFF2.otf')
    cff2_path = get_temp_file_path()
    runner(CMD + ['-a', '-o', 'cff2', '-f', font_path, cff2_path])
    expected_path = get_expected_path('regular_CFF2.cff2')
    assert differ([expected_path, cff2_path, '-m', 'bin'])


# -----------
# Other tests
# -----------

def test_trademark_string_pr425():
    # the copyright symbol used in the trademark field of a UFO is
    # converted to 'Copyright' and stored in Notice field of a Type1
    actual_path = runner(CMD + ['-s', '-o', 't1', '-f', 'trademark.ufo'])
    expected_path = get_expected_path('trademark.pfa')
    assert differ([expected_path, actual_path])


def test_remove_hints_bug180():
    font_path = get_input_path('cid.otf')
    cid_path = get_temp_file_path()
    runner(CMD + ['-a', '-o', 't1', 'n', '-f', font_path, cid_path])
    expected_path = get_expected_path('cid_nohints.ps')
    assert differ([expected_path, cid_path, '-m', 'bin'])


def test_long_charstring_read_bug444():
    # read a CFF2 VF with a charstring longer that 65535, check output
    actual_path = runner(CMD + ['-s', '-o', '0', '-f', 'CJK-VarTest.otf'])
    expected_path = get_expected_path('CJK-VarTest_read.txt')
    assert differ([expected_path, actual_path, '-s', '## Filename'])


def test_long_charstring_warning():
    # read a CFF2 VF with a charstring longer that 65535, check warning message
    # NOTE: can't diff the output against 'CJK-VarTest_warn.txt' because on
    # Windows the lines start with 'tx.exe:' instead of just 'tx:'
    actual_path = runner(
        CMD + ['-s', '-e', '-o', '5', '-f', 'CJK-VarTest.otf'])
    # expected_path = get_expected_path('CJK-VarTest_warn.txt')
    with open(actual_path, 'rb') as f:
        output = f.read()
    assert b"(cfr) Warning: CharString of GID 1 is 71057 bytes long" in output


def test_long_charstring_write():
    # read a CFF2 VF with a charstring longer that 65535, write out CFF2 file
    # NOTE: the font 'CJK-VarTest.otf' cannot be used in this test because
    # once its long charstring is optimized (floats -> ints) it's no longer
    # over the 65535 bytes limit; the long charstring in 'CJK-VarTest2.otf' is
    # already as small as possible, so it will trigger the check in cffwrite.c
    font_path = get_input_path('CJK-VarTest2.otf')
    cff2_path = get_temp_file_path()
    runner(CMD + ['-a', '-o', 'cff2', '-f', font_path, cff2_path])
    expected_path = get_expected_path('CJK-VarTest2.cff2')
    assert differ([expected_path, cff2_path, '-m', 'bin'])


def test_many_hints_string_bug354():
    # The glyph T@gid002 has 33 hstem hints. This tests a bug where
    # tx defined an array of only 6 operands.
    # This is encountered only when wrinting to a VF CFF2.
    font_path = get_input_path('cff2_vf.otf')
    cff2_path = get_temp_file_path()
    dcf_path = get_temp_file_path()
    runner(CMD + ['-a', '-o', 'cff2', '-f', font_path, cff2_path])
    runner(CMD + ['-a', '-o', 'dcf', '-f', cff2_path, dcf_path])
    expected_path = get_expected_path('cff2_vf.dcf.txt')
    assert differ([expected_path, dcf_path])


def test_non_varying_glyphs_bug356():
    """A glyph which is non-varying in a variable font may be referenced by a
    VariationStore data item subtable which has a region count of 0. The VF
    support code assumed that this was an error, and issued a false warning.
    File 'bug356.otf' is a handcrafted modification of 'cff2_vf.otf'. The
    latter cannot be used as-is to validate the fix."""
    actual_path = get_temp_file_path()
    font_path = get_input_path('bug356.otf')
    stderr_path = runner(CMD + ['-s', '-e', '-a', '-o', 'cff',
                                '-f', font_path, actual_path])
    expected_path = get_expected_path('bug356.txt')
    assert differ([expected_path, stderr_path, '-l', '1'])


@pytest.mark.parametrize('font_format', ['type1', 'cidfont', 'ufo2', 'ufo3'])
def test_no_psname_dump_bug437(font_format):
    if 'cid' in font_format:
        file_ext = 'ps'
    elif 'ufo' in font_format:
        file_ext = 'ufo'
    else:
        file_ext = 'pfa'

    filename = '{}-noPSname.{}'.format(font_format, file_ext)
    expected_path = get_expected_path('bug437/dump-{}.txt'.format(font_format))

    actual_path = runner(CMD + ['-s', '-o', 'dump', '0', '-f', filename])
    assert differ([expected_path, actual_path, '-l', '1'])


@pytest.mark.parametrize('font_format', ['type1', 'cidfont', 'ufo2', 'ufo3'])
def test_no_psname_convert_to_ufo_bug437(font_format):
    if 'cid' in font_format:
        file_ext = 'ps'
    elif 'ufo' in font_format:
        file_ext = 'ufo'
    else:
        file_ext = 'pfa'

    font_path = get_input_path('{}-noPSname.{}'.format(font_format, file_ext))
    expected_path = get_expected_path('bug437/{}.ufo'.format(font_format))
    save_path = os.path.join(
        _get_temp_dir_path(), '{}.ufo'.format(font_format))

    runner(CMD + ['-a', '-o', 'ufo', '-f', font_path, save_path])
    assert differ([expected_path, save_path])


@pytest.mark.parametrize('font_format', ['type1', 'cidfont', 'ufo2', 'ufo3'])
def test_no_psname_convert_to_type1_bug437(font_format):
    if 'cid' in font_format:
        file_ext = 'ps'
    elif 'ufo' in font_format:
        file_ext = 'ufo'
    else:
        file_ext = 'pfa'

    filename = '{}-noPSname.{}'.format(font_format, file_ext)
    with pytest.raises(subprocess.CalledProcessError) as err:
        runner(CMD + ['-o', 't1', '-f', filename])
    assert err.value.returncode in (5, 6)


def test_illegal_chars_in_glyph_name_bug473():
    font_path = get_input_path('bug473.ufo')
    save_path = os.path.join(_get_temp_dir_path(), 'bug473.ufo')
    runner(CMD + ['-a', '-o', 'ufo', '-f', font_path, save_path])
    expected_path = get_expected_path('bug473.ufo')
    assert differ([expected_path, save_path])


def test_subroutine_sorting_bug494():
    """ The input file was made with the command:
    tx -t1 -g 0-5 \
        source-serif-pro/Roman/Instances/Regular/font.ufo bug494.pfa
    The bug is that two subroutines in the Windows CFF output are swapped in
    index order from the Mac version. This was because of an unstable
    'qsort' done on the subroutines in the final stage of selection."""
    font_path = get_input_path('bug494.pfa')
    cff_path = get_temp_file_path()
    dcf_path = get_temp_file_path()
    runner(CMD + ['-a', '-o', 'cff', '*S', 'std', '*b',
                  '-f', font_path, cff_path])
    runner(CMD + ['-a', '-o', 'dcf', '-f', cff_path, dcf_path])
    expected_path = get_expected_path('bug494.dcf.txt')
    assert differ([expected_path, dcf_path])


@pytest.mark.parametrize('args, exp_filename', [([], 'roundtrip'),
                                                (['g', '_0-1'], 'subset')])
@pytest.mark.parametrize('to_format', ['t1', 'cff', 'afm'])
def test_recalculate_font_bbox_bug618(to_format, args, exp_filename):
    font_path = get_input_path('bug618.pfa')
    save_path = get_temp_file_path()

    runner(CMD + ['-f', font_path, save_path, '-o', to_format] + args)

    file_ext = to_format
    if to_format == 't1':
        file_ext = 'pfa'
    elif to_format == 'afm':
        file_ext = 'txt'

    expected_path = get_expected_path(
        'bug618/{}.{}'.format(exp_filename, file_ext))

    diff_mode = []
    if to_format == 'cff':
        diff_mode = ['-m', 'bin']

    skip = []
    if to_format == 'afm':
        skip = ['-s', 'Comment Creation Date:' + SPLIT_MARKER +
                'Comment Copyright']

    assert differ([expected_path, save_path] + diff_mode + skip)


def test_glyph_bboxes_bug655():
    actual_path = runner(CMD + ['-s', '-o', 'mtx', '2', '-f', 'bug655.ufo'])
    expected_path = get_expected_path('bug655.txt')
    assert differ([expected_path, actual_path])


@pytest.mark.parametrize('filename', ['SHSVF_9b3b', 'bug684'])
def test_cs_opt_bug684(filename):
    """ The input CFF2 variable font contains a long single charstring
    making the maximum use of the operand stack.
    tx was generating a bad CFF2 charstring that would overflow
    the operand stack of the standard size (513) after re-converted
    to CFF2 unless -no_opt option is specified."""
    font_path = get_input_path('{}.otf'.format(filename))
    result_path = get_temp_file_path()
    expected_path = get_expected_path('{}.cff2'.format(filename))
    runner(CMD + ['-a', '-o', 'cff2', '-f', font_path, result_path])
    assert differ([expected_path, result_path, '-m', 'bin'])


def test_standard_apple_glyph_names():
    actual_path = runner(CMD + ['-s', '-o', 'dump', '4', '-f', 'post-v2.ttf'])
    expected_path = get_expected_path('post-v2.txt')
    assert differ([expected_path, actual_path])


def test_ufo_self_closing_dict_element_bug701():
    actual_path = runner(CMD + ['-s', '-o', 'dump', '0', '-f', 'bug701.ufo'])
    expected_path = get_expected_path('bug701.txt')
    assert differ([expected_path, actual_path, '-s', '## Filename'])


def test_ufo3_guideline_bug705():
    actual_path = runner(CMD + ['-s', '-o', 't1', '-f', 'bug705.ufo'])
    expected_path = get_expected_path('bug705.pfa')
    assert differ([expected_path, actual_path])

def test_ufo_vertical_advance_bug786():
    actual_path = runner(CMD + ['-s', '-o', 't1', '-f', 'bug786.ufo'])
    expected_path = get_expected_path('bug786.pfa')
    assert differ([expected_path, actual_path])


@pytest.mark.parametrize('filename', [
    'a',  # AE glyph in both default and processed layers
    'b',  # AE glyph in default layer only
    'c',  # AE glyph in processed layer only
])
def test_ufo_read_processed_contents_plist_bug740(filename):
    actual_path = runner(CMD + ['-s', '-o', 'dump', '6', 'g', '_AE',
                                '-f', 'bug740/{}.ufo'.format(filename)])
    expected_path = get_expected_path('bug740/{}.txt'.format(filename))
    assert differ([expected_path, actual_path])



from __future__ import print_function, division, absolute_import

import os
import pytest

from fontTools.misc.py23 import tobytes

from runner import main as runner
from differ import main as differ
from test_utils import get_expected_path, get_temp_file_path

TOOL = 'otf2otc'
CMD = ['-t', TOOL]

REGULAR = 'SourceSansPro-Regular.otf'
ITALIC = 'SourceSansPro-It.otf'
BOLD = 'SourceSansPro-Bold.otf'

MSG_1 = tobytes(
    "Shared tables: "
    "['BASE', 'DSIG', 'GDEF', 'GSUB', 'cmap', 'maxp', 'post']%s"
    "Un-shared tables: "
    "['CFF ', 'GPOS', 'OS/2', 'head', 'hhea', 'hmtx', 'name']" % os.linesep)

MSG_2 = tobytes(
    "Shared tables: ['BASE', 'DSIG']%s"
    "Un-shared tables: ['CFF ', 'GDEF', 'GPOS', 'GSUB', 'OS/2', 'cmap', "
    "'head', 'hhea', 'hmtx', 'maxp', 'name', 'post']" % os.linesep)


# -----
# Tests
# -----

@pytest.mark.parametrize('otf_filenames, ttc_filename, tables_msg', [
    ([REGULAR, BOLD], 'RgBd.ttc', MSG_1),
    ([REGULAR, ITALIC], 'RgIt.ttc', MSG_2),
    ([REGULAR, ITALIC, BOLD], 'RgItBd.ttc', MSG_2),
])
def test_convert(otf_filenames, ttc_filename, tables_msg):
    actual_path = get_temp_file_path()
    expected_path = get_expected_path(ttc_filename)
    stdout_path = runner(CMD + ['-s', '-o', 'o', '_{}'.format(actual_path),
                                '-f'] + otf_filenames)
    with open(stdout_path, 'rb') as f:
        output = f.read()
    assert tables_msg in output
    assert differ([expected_path, actual_path, '-m', 'bin'])
